# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import io
import json
import ijson
import datetime
from collections.abc import Mapping, Iterable
from base64 import b64encode, b64decode
from decimal import Decimal
from uuid import UUID
import logging

logger = logging.getLogger('sesamclient.entity_json')


def datetime_parse(dt_str):
    if len(dt_str) <= 27:  # len('2015-11-24T07:58:53.123456Z') == 27
        if len(dt_str) == 10:  # len('2015-11-24') == 10
            dt = datetime.datetime.strptime(dt_str, "%Y-%m-%d")
        elif len(dt_str) <= 20:  # len('2015-11-24T07:58:53Z') == 27
            dt = datetime.datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%SZ")
        else:
            dt = datetime.datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%fZ")
    elif len(dt_str) > 30:  # len('2015-07-28T09:46:00.123456789Z') == 30
        raise Exception("Invalid date string: %s" % dt_str)
    else:
        dt = datetime.datetime.strptime(dt_str[:19], "%Y-%m-%dT%H:%M:%S")  # len('2015-11-24T07:58:53') == 19
        dt_str_digits = dt_str[19+1:-1]  # get number between . and Z
        dt_str_nanos = dt_str_digits.ljust(9, "0")
        nanos = int(dt_str_nanos)
        nanodelta = datetime.timedelta(microseconds=nanos/1000)
        dt = dt + nanodelta

    if dt.tzinfo is not None:
        dt = dt.astimezone(datetime.timezone.utc)
        dt = dt.replace(tzinfo=None)
    return dt

# -- JSON encoder

JSON_ENCODER_LAMBDAS = {
    #URI: lambda o: "~r%s" % o.value,
    bytes: lambda o: "~b%s" % str(b64encode(o), encoding="utf-8"),
    Decimal: lambda o: "~f%s" % o,
    float: lambda o: "~d%s" % o,
    UUID: lambda o: "~u%s" % str(o),
    datetime.date: lambda o: "~t%s" % o.isoformat(),
    datetime.datetime: lambda o: "~t%s" % o.isoformat() + "Z"
}

def _entity_json_encoder(o):
    """JSON encoder function. Usage: json.dumps(entity, default=_entity_json_encoder)"""
    value_type = type(o)
    json_encoder = JSON_ENCODER_LAMBDAS.get(value_type)
    if json_encoder is None:
        if isinstance(o, Mapping):
            return dict(o)
        elif not isinstance(o, str) and isinstance(o, Iterable):
            return list(o)
        else:
            raise TypeError(repr(o) + " is not JSON serializable")
    else:
        return json_encoder(o)

def entities_to_json(entities, sort_keys=False, indent=None, cls=None):
    """Returns a JSON serialized string from the given entities. This method is able
    to properly serialize non-native JSON types supported in entity bytes."""
    return json.dumps(entities, default=_entity_json_encoder, sort_keys=sort_keys,
                      indent=indent, cls=cls)

def parse_entity(entity_as_string, do_transit_decoding=True):
    dummy_stream_data = "[" + entity_as_string + "]"
    dummy_stream = io.BytesIO(dummy_stream_data.encode("utf-8"))
    entities = list(parse_json_stream(dummy_stream, do_transit_decoding=do_transit_decoding))
    assert len(entities) == 1
    return entities[0]


def parse_json_stream(stream, do_transit_decoding=True):
    """Parses stream of JSON entities. Stream can contain either a single
    JSON object or a list of JSON objects.
    :param stream: file-like stream object
    :param do_transit_decoding: bool determining if transit decoding should be performed
    :return: a generator of entity objects
    """
    context = []
    name_context = []
    for event, value in ijson.basic_parse(stream):
        if event == 'start_map':
            context.append({})
        elif event == 'map_key':
            name_context.append(value)
        elif event == 'string':
            ctxobj = context[-1]
            if type(ctxobj) is list:
                if do_transit_decoding is True and len(value) > 1 and value[0] == "~":
                    value1 = value[1]
                    #if value1 == "r":
                    #    ctxobj.append(URI(value[2:]))
                    if value1 == "t":
                        ctxobj.append(datetime_parse(value[2:]))
                    elif value1 == "b":
                        ctxobj.append(b64decode(value[2:]))
                    elif value1 == "u":
                        ctxobj.append(UUID(value[2:]))
                    elif value1 == "f":
                        ctxobj.append(Decimal(value[2:]))
                    elif value1 == "d":
                        ctxobj.append(float(value[2:]))
                    elif value1 == "~":
                        ctxobj.append(value[1:])
                    else:
                        ctxobj.append(value)
                else:
                    ctxobj.append(value)
            elif type(ctxobj) is dict:
                prop_name = name_context.pop()
                if do_transit_decoding is True and len(value) > 1 and value[0] == "~":
                    value1 = value[1]
                    #if value1 == "r":
                    #    ctxobj[prop_name] = URI(value[2:])
                    if value1 == "t":
                        ctxobj[prop_name] = datetime_parse(value[2:])
                    elif value1 == "b":
                        ctxobj[prop_name] = b64decode(value[2:])
                    elif value1 == "u":
                        ctxobj[prop_name] = UUID(value[2:])
                    elif value1 == "f":
                        ctxobj[prop_name] = Decimal(value[2:])
                    elif value1 == "d":
                        ctxobj[prop_name] = float(value[2:])
                    elif value1 == "~":
                        ctxobj[prop_name] = value[1:]
                    else:
                        ctxobj[prop_name] = value
                else:
                    ctxobj[prop_name] = value
            else:
                raise Exception("WAT!")
        elif event in {'number', 'boolean', 'null'}:
            ctxobj = context[-1]
            if type(ctxobj) is list:
                ctxobj.append(value)
            elif type(ctxobj) is dict:
                prop_name = name_context.pop()
                ctxobj[prop_name] = value
            else:
                raise Exception("WAT!")
        elif event == 'end_map':
            entity = context.pop()
            if len(context) == 1 and type(context[0]) is list or len(context) == 0:  # allow reading a single entity
                yield entity
            else:
                parent = context[-1]
                if type(parent) is dict:
                    parent[name_context.pop()] = entity
                elif type(parent) is list:
                    parent.append(entity)
                else:
                    raise Exception("WAT!")
        elif event == 'start_array':
            context.append([])
        elif event == 'end_array':
            l = context.pop()
            if len(context) > 0:
                parent = context[-1]
                if type(parent) is list:
                    parent.append(l)
                else:
                    parent[name_context.pop()] = l



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.

import shutil
import sys

from ..utils import print_item_list, exit_with_error


def get_logs_cmd(connection, args):
    # TODO: filtering
    logs = connection.get_logs()
    print_item_list(logs, args)


def get_log_content_cmd(connection, args):
    log_content = connection.get_log_content(args.log_id)
    if log_content is None:
        exit_with_error("No log with id '%s' was found!" % (args.log_id,))
    shutil.copyfileobj(log_content, sys.stdout.buffer)
    sys.stdout.flush()



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import os
import json

from ..utils import print_item, exit_with_error, print_item_list


def get_secrets_cmd(connection, args):
    system_id = args.system_id
    if system_id:
        system = connection.get_system(system_id)
        if system is None:
            exit_with_error("No system with id '%s' was found!" % (args.system_id,))
        secrets_info = system.get_secrets()
    else:
        secrets_info = connection.get_secrets()
    print_item_list(secrets_info, args)


def put_secrets_cmd(connection, args):
    secrets_path = args.secrets_file

    if not os.path.exists(secrets_path):
        exit_with_error("The specified path '%s' doesn't exist!" % (secrets_path,))
    with open(secrets_path, "r", encoding="utf-8") as secrets_content:
        secrets = json.loads(secrets_content.read())

    system_id = args.system_id
    if system_id:
        system = connection.get_system(system_id)
        if system is None:
            exit_with_error("No system with id '%s' was found!" % (args.system_id,))
        secrets_info = system.put_secrets(secrets, args.dont_encrypt)
    else:
        secrets_info = connection.put_secrets(secrets, args.dont_encrypt)
    print_item(secrets_info, args)


def post_secrets_cmd(connection, args):
    secrets_path = args.secrets_file

    if not os.path.exists(secrets_path):
        exit_with_error("The specified path '%s' doesn't exist!" % (secrets_path,))
    with open(secrets_path, "r", encoding="utf-8") as secrets_content:
        secrets = json.loads(secrets_content.read())

    system_id = args.system_id
    if system_id:
        system = connection.get_system(system_id)
        if system is None:
            exit_with_error("No system with id '%s' was found!" % (args.system_id,))
        secrets_info = system.post_secrets(secrets, args.dont_encrypt)
    else:
        secrets_info = connection.post_secrets(secrets, args.dont_encrypt)
    print_item(secrets_info, args)


def delete_secret_cmd(connection, args):
    system_id = args.system_id
    if system_id:
        system = connection.get_system(system_id)
        if system is None:
            exit_with_error("No system with id '%s' was found!" % (args.system_id,))
        secret_info = system.delete_secret(args.secret_key)
    else:
        secret_info = connection.delete_secret(args.secret_key)
    print_item(secret_info, args)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import pprint
import textwrap
import time
import copy
import dateutil.parser

from .entitybase import EntityBase
from .exceptions import PumpIsAlreadyRunning, TimeoutWhileWaitingForRunningPumpToFinishException, \
    PumpDoesNotSupportTheOperation
from .utils import validate_response_is_ok


def parse_datetime(datetimestring):
    if not datetimestring:
        return None
    return dateutil.parser.parse(datetimestring)


class Pump(EntityBase):
    """
    """

    def __init__(self, pipe, connection, raw_jsondata):
        self.pipe = pipe
        self.pipe_id = pipe.id
        super().__init__(connection, raw_jsondata)

    @property
    def id(self):
        return "pump:" + self.pipe_id

    @property
    def type(self):
        return "pump"

    @property
    def runtime(self):
        return copy.deepcopy(self._raw_jsondata["runtime"])

    def __str__(self):
        result = super().__str__()
        result += "\n    pipe_id:%r" % (self.pipe_id,)
        return result

    def get_is_running(self, **kwargs):
        # Note: we must retrieve up-to-date information from the server, since the runtime stuff is so
        #       dynamic.
        self.update_raw_jsondata_from_url(self._connection.get_pipe_pump_url(self.pipe_id), **kwargs)
        return self._raw_jsondata["runtime"]["is-running"]

    def update_raw_jsondata(self, new_raw_jsondata):
        super().update_raw_jsondata(new_raw_jsondata)

        # Patch up the _raw_jsondata on the pipe, as well.
        # noinspection PyProtectedMember
        if "runtime" in self._raw_jsondata:
            self.pipe._raw_jsondata["runtime"].update(self._raw_jsondata["runtime"])

    def wait_for_pump_to_finish_running(self, timeout=60, poll_interval=0.5, verbosity=0, **kwargs):
        """Utility method for waiting until the pump is no longer running. This is useful in
        integration tests that want to make sure that the pump has been run after the test-code has
        modified the data in the source-system of a pipe.

        :param timeout: The maximum time to wait for the pump to stop running. If the timeout is exceeded,
                        an error is raised.
        """
        starttime = time.monotonic()
        time_of_last_log_output = 0
        while True:
            if not self.get_is_running():
                # ok, we are not running, so we can exit now.
                break

            elapsedtime = time.monotonic() - starttime
            if elapsedtime > timeout:
                self.pipe.update_raw_jsondata_from_url(self._connection.get_pipe_url(self.pipe.id), **kwargs)
                raise TimeoutWhileWaitingForRunningPumpToFinishException(
                    "Timed out while waiting for the pump to finish running! pipe:\n%s" % (
                        self.pipe.as_json(),
                    ))

            if verbosity > 0:
                current_time = time.monotonic()
                if (time.monotonic() - time_of_last_log_output) > 3:
                    if verbosity > 1:
                        print("Waiting for the pipe '%s' to finish running (elapsed time: %.2fs). "
                              "Progress:\n%s\n  Last-seen:%s" % (
                                  self.id,
                                  elapsedtime,
                                  textwrap.indent(pprint.pformat(self.runtime["progress"]), prefix="    "),
                                  self.runtime.get("last-seen"),
                              ))
                    else:
                        print("Waiting for the pipe '%s' to finish running (elapsed time: %.2fs). "
                              "Processed:%s  Last-seen:%s" % (
                                  self.id,
                                  elapsedtime,
                                  self.runtime["progress"]["processed"],
                                  self.runtime.get("last-seen"),
                              ))
                    time_of_last_log_output = current_time

            time.sleep(poll_interval)

    def has_pending_work(self, **kwargs):
        """Utility method for checking that the pumps source-queue and dependency-queue are empty. If the
        queues are empty, the pump will not produce any output if it is run.
        This is useful in integration tests that want to make sure that all indexing, etc that the pump
        depends on has been done, so that the pump output has reached its "eventually consistent" state.
        """
        if self.get_is_running(**kwargs):
            # The pump is running, so clearly it has pending work
            return True

        # Note: the get_is_running() method will have retrieved up-to-date information from the server,
        #       so we don't have to do it here.
        queues = self._raw_jsondata["runtime"]["queues"]

        def has_pending_work(dict_or_integer):
            if dict_or_integer is None:
                return False

            if isinstance(dict_or_integer, dict):
                for dataset_id, queue_size in dict_or_integer.items():
                    if queue_size is not None and queue_size > 0:
                        return True
            else:
                if dict_or_integer > 0:
                    return True
            return False

        if has_pending_work(queues.get("dependencies")):
            return True

        if has_pending_work(queues.get("source")):
            return True

        return False

    def run_pump_until_there_are_no_pending_work(self, timeout=60, sleep_time_between_runs=0.5, verbosity=0, **kwargs):
        """Utility method for running the pump until the has_pending_work() method return False.

        :param timeout: The maximum time to wait for the pump to finish. If the timeout is exceeded,
                        an error is raised.
        :sleep_time_between_runs: How long to sleep before restarting the pipe when the has_pending_work() method
                                  has returned True.

        """
        self.wait_for_pump_to_finish_running(timeout=timeout, verbosity=verbosity, **kwargs)
        self.start(**kwargs)
        self.wait_for_pump_to_finish_running(timeout=timeout, verbosity=verbosity, **kwargs)

        starttime = time.monotonic()
        while True:
            if not self.has_pending_work(**kwargs):
                # Ok, we are done.
                return

            elapsedtime = time.monotonic() - starttime
            if elapsedtime > timeout:
                raise TimeoutWhileWaitingForRunningPumpToFinishException(
                    "Timed out while waiting for the '%s' pump to complete all its pending work! pump:\n%s" % (
                        self.pipe_id, self.as_json(),))
            time.sleep(sleep_time_between_runs)

            self.wait_for_pump_to_finish_running(timeout=timeout, verbosity=verbosity, **kwargs)
            self.start(**kwargs)
            self.wait_for_pump_to_finish_running(timeout=timeout, verbosity=verbosity, **kwargs)

    @property
    def last_message(self):
        return self._raw_jsondata["runtime"]["last-message"]

    @property
    def last_original_error_message(self):
        return self._raw_jsondata["runtime"].get("last-original-error-message")

    @property
    def last_seen(self):
        return self._raw_jsondata["runtime"]["last-seen"]

    def reset_tracking(self, **kwargs):
        self.run_operation("reset-tracking", allowable_response_status_codes=frozenset([200]), **kwargs)

    def set_tracking_to_start(self, **kwargs):
        self.run_operation("set-tracking-to-start", allowable_response_status_codes=frozenset([200]), **kwargs)

    def unset_last_seen(self, **kwargs):
        self.update_last_seen("", **kwargs)

    def set_last_seen_to_end(self, **kwargs):
        self.run_operation("set-last-seen-to-end", allowable_response_status_codes=frozenset([200]), **kwargs)

    def update_last_seen(self, last_seen, **kwargs):
        last_seen = str(last_seen)
        self.run_operation("update-last-seen", operation_parameters={"last-seen": last_seen},
                           allowable_response_status_codes=frozenset([200]), **kwargs)

        # Check that the "last-seen"-attribute has been updated
        actual_last_seen = self.last_seen
        if actual_last_seen != last_seen:
            if (last_seen is None and actual_last_seen == "") or (last_seen == "" and actual_last_seen is None):
                # None and an empty string means the same here.
                pass
            else:
                raise AssertionError(
                    "Failed to update the 'last-seen'-value of the pump '%s'! expected value: '%s'  actual value: '%s'" % (
                        self.id, last_seen, actual_last_seen))

    @property
    def is_disabled(self):
        return self._raw_jsondata["runtime"]["is-disabled"]

    @property
    def is_valid_config(self):
        return self._raw_jsondata["runtime"]["is-valid-config"]

    @property
    def config_errors(self):
        return self._raw_jsondata["runtime"]["config-errors"]

    @property
    def last_run(self):
        return parse_datetime(self._raw_jsondata["runtime"]["last-run"])

    @property
    def next_run(self):
        return parse_datetime(self._raw_jsondata["runtime"]["next-run"])

    def disable(self, comment=None, **kwargs):
        """
        Disable this pump. This prevents the pump from being automatically run by the task scheduler.
        Can add comment parameter that shows up in execution log.
        """
        self._set_disabled_state(True, comment=comment, **kwargs)

    def enable(self, **kwargs):
        """Enable this pump."""
        self._set_disabled_state(False, **kwargs)

    def _set_disabled_state(self, is_disabled, comment=None, **kwargs):
        # Send a put-request to disable the pump
        if self.is_disabled != is_disabled:
            if is_disabled:
                operation = "disable"
            else:
                operation = "enable"

            if operation == "disable" and comment is not None:
                operation_parameters = { comment: comment}
            else:
                operation_parameters = None

            self.run_operation(operation, operation_parameters=operation_parameters,
                                allowable_response_status_codes=frozenset([200]), **kwargs)

            # Check that the "is-disabled"-attribute has been updated
            actual_is_disabled = self.is_disabled
            if actual_is_disabled != is_disabled:
                raise AssertionError(("Failed to update the 'is-disabled' setting on the pump '"
                                      "'%s'! Expected value: '%s'  Actual value: '%s'") % (self.id, is_disabled,
                                                                                           actual_is_disabled))

    def start(self, dont_wait_for_pump_to_start=False, allow_already_running_pump=True, operation_parameters={},
              **kwargs):
        """
        Starts the pump.

        :param dont_wait_for_pump_to_start: If this flag is set to True, the method will return without waiting for the
                                           pump to start running.
        :param allow_already_running_pump: If this  flag is set to False, this method will fail with a
                                           "PumpIsAlreadyRunning"-exception if if the pump was already
                                           running.
        """
        if dont_wait_for_pump_to_start:
            operation_parameters["dont-wait-for-pump-to-start"] = "true"

        response = self.run_operation("start", operation_parameters=operation_parameters,
                                      allowable_response_status_codes=frozenset([200, 409]), **kwargs)

        if response.status_code == 409:
            self.update_raw_jsondata(response.content.decode("utf-8"))
            if allow_already_running_pump:
                return False
            else:
                raise PumpIsAlreadyRunning(self)

        validate_response_is_ok(response, 200)
        self.update_raw_jsondata(response.content.decode("utf-8"))
        return True

    def stop(self, **kwargs):
        """
        Stops the pump.
        """
        self.run_operation("stop", **kwargs)

    def start_rescan(self, allow_already_running_pump=True, **kwargs):
        """
        Starts a rescan-run.
        :param allow_already_running_pump: If this  flag is set to False, this method will fail with a
                                           "PumpIsAlreadyRunning"-exception if if the pump was already
                                           running.
        """
        response = self.run_operation("start-rescan", allowable_response_status_codes=frozenset([200, 409]), **kwargs)

        if response.status_code == 409:
            self.update_raw_jsondata(response.content.decode("utf-8"))
            if allow_already_running_pump:
                return False
            else:
                raise PumpIsAlreadyRunning(self)

        validate_response_is_ok(response, 200)
        self.update_raw_jsondata(response.content.decode("utf-8"))
        return True

    def reset_rescan(self, **kwargs):
        """
        Resets the rescan-run state.
        """
        response = self.run_operation("reset-rescan", allowable_response_status_codes=frozenset([200]), **kwargs)
        validate_response_is_ok(response, 200)
        self.update_raw_jsondata(response.content.decode("utf-8"))
        return True

    @property
    def supported_operations(self):
        return self._raw_jsondata["runtime"]["supported-operations"]

    def run_operation(self,
                      operation,
                      operation_parameters=None,
                      allowable_response_status_codes=frozenset([200]),
                      **kwargs):
        """Runs the specified operation on the pump.
        :param operation: The operation to run.
                          Must be one of the values of the :py:meth:`supported_operations` property.
        :param operation_parameters: A dict with operation parameters.
        :param allowable_response_status_codes: The http request status code(s) to accept (defaults to 200).
        """
        url = self._connection.get_pipe_pump_url(self.pipe_id)
        postdata = {"operation": operation}
        if operation_parameters is not None:
            postdata.update(operation_parameters)

        _allowable_response_status_codes = list(allowable_response_status_codes).append(400)
        response = self._connection.do_post_request(url,
                                                    allowable_response_status_codes=_allowable_response_status_codes,
                                                    data=postdata, **kwargs)

        # Handle the "operation not supported" 400 response, unless we have been told that 400-responses is
        # ok (in which case it is the callers responsibility to handle the error).
        try:
            response_text = response.content.decode("utf-8")
        except:
            # In case it isn't utf-8
            response_text = response.text

        if response.status_code == 400 and 400 not in allowable_response_status_codes:
            # check if this was caused by the pump not supporting the requested operation
            #
            if (
                    "doesn&#x27;t support the operation" in response_text or  # html response
                    "doesn't support the operation" in response_text):  # json response
                raise PumpDoesNotSupportTheOperation(pump=self,
                                                     operation=operation,
                                                     response=response)

        validate_response_is_ok(response, allowable_response_status_codes=allowable_response_status_codes)
        self.update_raw_jsondata(response_text)
        return response



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import copy
import io
import json
import logging
import pprint

from .entity_json import parse_json_stream
from .utils import validate_response_is_ok

logger = logging.getLogger(__name__)


class EntityBase:
    """
    This is the common baseclass for all the classes that represents various server-side objects (Pipe, System,
    Dataset, etc)
    """
    def __init__(self, connection, raw_jsondata):
        self._connection = connection
        self._raw_jsondata = {} # we initially set this to an empty dict since it is used by the __repr__()-method.
        if raw_jsondata is None:
            raise AssertionError("The raw_jsondata parameter was None! That makes no sense!")
        if not isinstance(raw_jsondata, dict):
            raise AssertionError(
                "The raw_jsondata isn't a dict! That makes no sense! type(raw_jsondata):'%s' raw_jsondata:'%s'" % (
                    type(raw_jsondata), raw_jsondata))
        self._raw_jsondata = raw_jsondata

    def update_raw_jsondata_from_url(self, url, **kwargs):
        """Utility-method for updating this object with new information from the specified url. This is typically
        used by a subclass when it knows that an operation has resulted in a change in the state, or when it knows
        that it needs up-to-date information (example: ScheduledTash.get_running_task())
        """
        response = self._connection.do_get_request(url, allowable_response_status_codes=[200], **kwargs)
        self.update_raw_jsondata(response.content.decode("utf-8"))

    def update_raw_jsondata(self, new_raw_jsondata):
        if isinstance(new_raw_jsondata, str):
            new_raw_jsondata = json.loads(new_raw_jsondata)
        self._raw_jsondata = new_raw_jsondata

    def __repr__(self):
        return "%(typename)s(id=%(id)r)" % dict(
            typename=self.type,
            id=self.id,
        )

    @property
    def id(self):
        theid = self._raw_jsondata.get("_id") or self._raw_jsondata.get("id")
        if not theid:
            theid = "*no id found! Known info: %s*" % (self.get_debug_info(), )
        return theid

    @property
    def subjectidentifier(self):
        return self._raw_jsondata.get("subjectidentifier")

    @property
    def type(self):
        thetype = self._raw_jsondata.get("type")
        if thetype:
            return thetype
        return self.config.get("type", self.__class__.__name__)

    @property
    def config(self):
        return self.get_config()

    def get_config(self, do_transit_decoding=False):
        return self.as_json_object(do_transit_decoding=do_transit_decoding).get("config", {})

    def as_json(self):
        """Returns a json string representation of this object"""
        return json.dumps(self._raw_jsondata, indent=4)

    def as_json_object(self, do_transit_decoding=False):
        """Returns a json object representation of this object"""
        if do_transit_decoding:
            dummy_stream = io.BytesIO(json.dumps([self._raw_jsondata]).encode("utf-8"))
            entities = list(parse_json_stream(dummy_stream, do_transit_decoding=do_transit_decoding))
            assert len(entities) == 1
            jsondata_with_transit_decoded_values = entities[0]
            return jsondata_with_transit_decoded_values
        else:
            return copy.deepcopy(self._raw_jsondata)

    def get_debug_info(self):
        return "self._raw_jsondata:%s" % (pprint.pformat(self._raw_jsondata),)

    @property
    def cached_response_info(self):
        return self._raw_jsondata.get("cached_response_info")



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.

import json
import logging
import copy

from .exceptions import ConfigUploadFailed
from .entitybase import EntityBase
from . import utils


logger = logging.getLogger(__name__)


class Model(EntityBase):
    """
    This class represents a model.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def delete(self, **kwargs):
        """Deletes this system from the sesam-node.
        """
        url = self._connection.get_model_url(self.id)
        response = self._connection.do_delete_request(url, allowable_response_status_codes=[200],
                                                      **kwargs)
        self.update_raw_jsondata(response.content.decode("utf-8"))

    @property
    def config(self):
        return self.get_config()

    def get_config(self, do_transit_decoding=False):
        return self.as_json_object(do_transit_decoding=do_transit_decoding)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import shutil

import sys

from ..utils import print_item, print_item_list, exit_with_error, print_message


def get_datasets_cmd(connection, args):
    # TODO: filtering
    datasets = connection.get_datasets()
    print_item_list(datasets, args)


def get_dataset_cmd(connection, args):
    dataset = connection.get_dataset(args.dataset_id)

    if dataset is None:
        exit_with_error("No dataset with id '%s' was found!" % (args.dataset_id,))

    if args.out == "json":
        print_message(dataset.as_json())
    else:
        print_message(str(dataset))


def delete_dataset_cmd(connection, args):
    dataset = connection.get_dataset(args.dataset_id)
    if dataset is None:
        exit_with_error("No dataset with id '%s' was found!" % (args.dataset_id,))
    dataset.delete()
    print_item(dataset, args)


def get_dataset_entities_cmd(connection, args):
    dataset = connection.get_dataset(args.dataset_id)

    if dataset is None:
        exit_with_error("No dataset with id '%s' was found!" % (args.dataset_id,))

    # We must take care when printing out the entities, since it could be billions of them. We can't make a list and
    # print everything in one go, since that could take too much resources. Instead we get the raw byte-stream
    # from the server and copy it to sys.stdout.buffer. This is orders of magnitude faster than using the
    # print_message() function (or sys.stdout.write()) for each line.
    entities_stream = dataset.get_entities_as_stream(since=args.since, limit=args.limit, start=args.start,
                                                     history=args.history, deleted=args.deleted)
    shutil.copyfileobj(entities_stream, sys.stdout.buffer)
    sys.stdout.flush()


def get_dataset_entity_cmd(connection, args):
    dataset = connection.get_dataset(args.dataset_id)
    entity_id = args.entity_id

    if dataset is None:
        exit_with_error("No dataset with id '%s' was found!" % (args.dataset_id,))

    entity = dataset.get_entity(entity_id)
    print_item(entity, args)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import pprint

from ..utils import print_item, print_item_list, exit_with_error
from ...exceptions import SystemAlreadyExists


def get_systems_cmd(connection, args):
    # TODO: filtering
    systems = connection.get_systems()
    print_item_list(systems, args)


def get_system_cmd(connection, args):
    system = connection.get_system(args.system_id)
    if system is None:
        exit_with_error("No system with id '%s' was found!" % (args.system_id,))
    print_item(system, args)


def add_systems_cmd(connection, args):
    try:
        systems = connection.add_systems(args.system_configurations, force=args.force)
        print_item_list(systems, args)
    except SystemAlreadyExists as e:
        exit_with_error("One or more of the systems already exist: %s" % (e,))


def modify_system_cmd(connection, args):
    system_config = args.system_configuration
    system_id = system_config.get("_id")
    if not system_id:
        exit_with_error("The system-configuration doesn't contain a '_id' attribute! system_config: \n%s" % (
            pprint.pformat(system_config),))

    system = connection.get_system(system_id)
    if system is None:
        exit_with_error("No system with id '%s' was found!" % (system_id,))

    system.modify(system_config, force=args.force)
    print_item(system, args)


def delete_system_cmd(connection, args):
    system = connection.get_system(args.system_id)
    if system is None:
        exit_with_error("No system with id '%s' was found!" % (args.system_id,))
    system.delete()
    print_item(system, args)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import argparse
import inspect
import json
import logging
import os.path
from html.parser import HTMLParser

import requests.exceptions
import sys

import time

from .commands.login import logout_cmd, login_cmd
from ..exceptions import PumpDoesNotSupportTheOperation, BadRequestException
from .commands.license import get_license_cmd, put_license_cmd
from .commands.config import config_cmd, get_configparser, get_configfilename
from .commands.datasets import delete_dataset_cmd, get_dataset_cmd, get_dataset_entities_cmd, \
    get_dataset_entity_cmd, get_datasets_cmd
from .commands.export import export_cmd
from .commands.import_cmd import import_cmd
from .commands.logs import get_log_content_cmd, get_logs_cmd
from .commands.secretsmanager import get_secrets_cmd, delete_secret_cmd, put_secrets_cmd, post_secrets_cmd
from .commands.envvarsmanager import get_env_vars_cmd, delete_env_var_cmd, put_env_vars_cmd, post_env_vars_cmd
from .commands.pipes import get_pipe_cmd, get_pipes_cmd, add_pipes_cmd, delete_pipe_cmd, \
    modify_pipe_cmd, get_pipe_schema_def_cmd
from .commands.pumps import disable_pump_cmd, enable_pump_cmd, get_pump_cmd, run_pump_operation_cmd, start_pump_cmd, \
    stop_pump_cmd, unset_last_seen_cmd, update_last_seen_cmd, wait_for_pump_to_finish_running_cmd
from .commands.systems import get_system_cmd, get_systems_cmd, add_systems_cmd, delete_system_cmd, modify_system_cmd
from .utils import exit_with_error
from .. import Connection
from .. import utils
from . import utils as main_utils
logger = logging.getLogger(__name__)


def get_parser_used_by_sphinx_argparse_extension():
    """This is the method that is used when autogeneration documentation with the sphinx argparse extension.
    It has a quirk (or bug?) that causes it to print an "Undocumented" message  for subparsers that don't have
    a "help"-parameter, even for subparsers that has a "description"-parameter. The "description"-parameter is the one
    that is actually displayed on the command line, so that is the one we want to use. But to get the sphinx-stuff to
    work we must specify the "help"-param instead of the "description"-param when we generate the parser the
    sphinx-stuff sees.
    """
    return get_parser(use_help_instead_of_description_for_subparsers=True)


def get_parser(use_help_instead_of_description_for_subparsers=False):
    parser_description = """
description:
    This is a command line tool for connecting to a Sesam installation and performing
    various operations.

    The root address of the Sesam installation rest api can be specified with the --"server_base_url"
    argument, or by by using the 'config' command to store the address persistently.

    To get detailed help for a command, add '-h' after the command, like this:
      %%(prog)s config -h

    (version: %s)
    """ % (utils.get_version(),)

    parser = argparse.ArgumentParser(description=parser_description,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument(
        "--loglevel",
        help="The loglevel to use. The default is 'WARNING'",
        choices=['ERROR', 'WARNING', 'INFO', 'DEBUG'],
        default='WARNING')

    # We use a separate subparser for each command. Each subparser gets assigned a function that gets called
    # when the user specifies the command in the command line arguments. This main() function does all the common
    # setup, while the subparser-functions does all the command-specific stuff.
    subparsers = parser.add_subparsers(dest="command")

    # noinspection PyShadowingBuiltins
    def add_subparser(*args,
                      add_default_arguments=True,
                      add_default_arguments_for_list_commands=False,
                      add_default_output_format_argument=True,
                      help=None, description=None,
                      **kwargs):
        subparser_kwargs = dict(**kwargs)
        subparser_kwargs["help"] = help
        subparser_kwargs["description"] = description
        if use_help_instead_of_description_for_subparsers and (help is None):
            subparser_kwargs["help"] = description
            subparser_kwargs["description"] = None
        subparser = subparsers.add_parser(*args, **subparser_kwargs)

        if add_default_arguments:

            subparser.add_argument(
                "--server_base_url",
                metavar="<server base url>",
                help="""The server base url to use. If specified, this overrides any default value that has been set with the
                        'config' command. Example usage: --server_base_url http://localhost:10001/api""")

            subparser.add_argument(
                "--jwt",
                metavar="<jwt filename>",
                help="""Specify a file containg a JWT authorization token to use when talking to the server. If
                        specified, this overrides any JWT-token that has been stored as a result of a 'sesam login'
                        command. Example usage: --jwt ~/my_jwt_tokens/prodserver1.jwt""")

            subparser.add_argument(
                "--server_username",
                metavar="<username>",
                help="""The username to use when authenticating to the server. If specified, this overrides any default value
                        that has been set with the
                        'config' command. Example usage: --server_username olanorman""")

            subparser.add_argument(
                "--server_password",
                metavar="<password>",
                help="""The password to use when authenticating to the server. If specified, this overrides any default value
                        that has been set with the
                        'config' command. Example usage: --server_password verysecret""")

            subparser.add_argument(
                "--server_timeout",
                metavar="<timeout in seconds>",
                help="""The server timeout to use. If specified, this overrides any default value that has been set with the
                        'config' command. Example usage: --server_timeout 60""")

        if add_default_arguments and add_default_output_format_argument:
            subparser.add_argument(
                "--out",
                choices=['json', 'yaml', 'plain'],
                default='json',
                help="""The output format to use. The default format is 'json'.""")

        if add_default_arguments_for_list_commands:
            subparser.add_argument(
                "--ids-only",
                default=False,
                action='store_true',
                help="""If this is specified, only the id's of the items will be displayed.""")

        return subparser

    def add_force_config_argument(subparser):
        subparser.add_argument(
            "--force",
            default=False,
            action="store_true",
            help="Forces the server to accept configuration errors and do a best-effort attempt to apply the new "
                 "configuration. This is rarely a good idea.")

    # The "config"-command is used to set or get persistently stored configuration options (server root url, etc).
    configcmd_parser = add_subparser(
        "config",
        description="Gets or sets a config option.",
        add_default_arguments=False  # The config sub-command is special, and doesn't use the default args.
    )
    configcmd_parser.add_argument(
        "optionname", help="The config option to get or set",
        choices=["server.base_url",
                 "server.username",
                 "server.password",
                 "server.timeout",
                 ])
    configcmd_parser.add_argument(
        "optionvalue",
        help="If specified, the config option will be set to this value. "
             "If no value is specified, the current value will be displayed.",
        nargs=argparse.REMAINDER,
        default=None)
    configcmd_parser.set_defaults(subparser_func=config_cmd)

    get_pipes_parser = add_subparser(
        "get-pipes",
        description="Display a list of pipes",
        add_default_arguments_for_list_commands=True)
    get_pipes_parser.set_defaults(subparser_func=get_pipes_cmd)

    get_pipe_parser = add_subparser(
        "get-pipe",
        description="Display detailed information about the specified pipe")
    get_pipe_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    get_pipe_parser.set_defaults(subparser_func=get_pipe_cmd)

    get_pipe_schema_def_parser = add_subparser(
        "get-pipe-schema-definition",
        description="Generate a schema defintion for the pipe")
    get_pipe_schema_def_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    get_pipe_schema_def_parser.add_argument(
        "sample_size",
        help="The number of entities to sample for the schema.", default=1000)
    get_pipe_schema_def_parser.add_argument(
        "--keys_only",
        default=False,
        action="store_true",
        help="If set only returns the keys of the schema as an array.")

    get_pipe_schema_def_parser.set_defaults(subparser_func=get_pipe_schema_def_cmd)

    add_pipes_parser = add_subparser(
        "add-pipes",
        description="Add one or more new pipes",
        add_default_arguments_for_list_commands=True
    )
    add_pipes_parser.add_argument(
        "pipe_configurations",
        type=json.loads,
        help="A list of the configurations of the new pipes. Each pipe configuration should be a json-encoded dict on "
             "the same format as used when  defining a pipe in the sesam-node configuration file. "
             """Example: '[{"_id": "my_new_pipe", "type": "pipe", "short_config": "sql://Northwind/Customers"}]'
             """)
    add_force_config_argument(add_pipes_parser)
    add_pipes_parser.set_defaults(subparser_func=add_pipes_cmd)

    modify_pipe_parser = add_subparser(
        "modify-pipe",
        description="Modify a pipe by uploading a new configuration for the pipe")
    modify_pipe_parser.add_argument(
        "pipe_configuration",
        type=json.loads,
        help="The new configuration of the pipe. This should be a json object on the same format as used when "
             "defining a pipe in the sesam-node configuration file. The pipe to modify is specified by the '_id' "
             "attribute of the configuration. "
             """Example: '{"_id": "my_existing_pipe", "type": "pipe", "short_config": "sql://Northwind/Customers"}'"
             """)
    add_force_config_argument(modify_pipe_parser)
    modify_pipe_parser.set_defaults(subparser_func=modify_pipe_cmd)

    delete_pipe_parser = add_subparser(
        "delete-pipe",
        description="Deletes the specified pipe from the sesam-node")
    delete_pipe_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    delete_pipe_parser.set_defaults(subparser_func=delete_pipe_cmd)

    get_logs_parser = add_subparser(
        "get-logs",
        description="Display a list of logs",
        add_default_arguments_for_list_commands=True)
    get_logs_parser.set_defaults(subparser_func=get_logs_cmd)

    get_log_parser = add_subparser(
        "get-log",
        description="Display the content of the specified log")
    get_log_parser.add_argument(
        "log_id",
        help="The id of the log.")
    get_log_parser.set_defaults(subparser_func=get_log_content_cmd)

    get_systems_parser = add_subparser(
        "get-systems",
        description="Display a list of systems",
        add_default_arguments_for_list_commands=True)
    get_systems_parser.set_defaults(subparser_func=get_systems_cmd)

    get_system_parser = add_subparser(
        "get-system",
        description="Display detailed information about the specified system")
    get_system_parser.add_argument(
        "system_id",
        help="The id of the system.")
    get_system_parser.set_defaults(subparser_func=get_system_cmd)

    add_systems_parser = add_subparser(
        "add-systems",
        description="Add one or more new systems",
        add_default_arguments_for_list_commands=True
    )
    add_systems_parser.add_argument(
        "system_configurations",
        type=json.loads,
        help="A list of the configurations of the new systems. Each system configuration should be a json-encoded "
             "dict on the same format as used when  defining a system in the sesam-node configuration file. "
             """Example: '[{"_id": "my_new_system", "type": "system", "short_config": "sql://Northwind/Customers"}]'
             """)
    add_force_config_argument(add_systems_parser)
    add_systems_parser.set_defaults(subparser_func=add_systems_cmd)

    modify_system_parser = add_subparser(
        "modify-system",
        description="Modify a system by uploading a new configuration for the system")
    modify_system_parser.add_argument(
        "system_configuration",
        type=json.loads,
        help="The new configuration of the system. This should be a json object on the same format as used when "
             "defining a system in the sesam-node configuration file. The system to modify is specified by the '_id' "
             "attribute of the configuration. "
             """Example: '{"_id": "my_existing_system", "type": "system", "short_config": "sql://Northwind/Customers"}'"
             """)
    add_force_config_argument(modify_system_parser)
    modify_system_parser.set_defaults(subparser_func=modify_system_cmd)

    delete_system_parser = add_subparser(
        "delete-system",
        description="Deletes the specified system from the sesam-node")
    delete_system_parser.add_argument(
        "system_id",
        help="The id of the system.")
    delete_system_parser.set_defaults(subparser_func=delete_system_cmd)

    get_datasets_parser = add_subparser(
        "get-datasets",
        description="Display a list of datasets",
        add_default_arguments_for_list_commands=True)
    get_datasets_parser.set_defaults(subparser_func=get_datasets_cmd)

    get_dataset_parser = add_subparser(
        "get-dataset",
        description="Display detailed information about the specified dataset")
    get_dataset_parser.add_argument(
        "dataset_id",
        help="The id of the dataset.")
    get_dataset_parser.set_defaults(subparser_func=get_dataset_cmd)

    delete_dataset_parser = add_subparser(
        "delete-dataset",
        description="Deletes the specified dataset from the sesam-node")
    delete_dataset_parser.add_argument(
        "dataset_id",
        help="The id of the dataset.")
    delete_dataset_parser.set_defaults(subparser_func=delete_dataset_cmd)

    get_dataset_entities_parser = add_subparser(
        "get-dataset-entities",
        description="Returns the entities of the specified dataset")
    get_dataset_entities_parser.add_argument(
        "dataset_id",
        help="The id of the dataset.")
    get_dataset_entities_parser.add_argument(
        "--since",
        help="""The "since" parameter is the opaque value that can be used to efficiently skip already \
seen entities.""", default=None)
    get_dataset_entities_parser.add_argument(
        "--limit",
        help="""The "limit" parameter specifies the maximum number of entities to return. \
If this is not specified, all available entities are returned.""",
        default=None)
    get_dataset_entities_parser.add_argument(
        "--start",
        help="""The "start" parameter is a positive integer that specified how many entities to skip from the \
start of where the "since"-parameters starts. This is far less efficient than using the "since"-parameter, but it \
can be manually set in a gui to explore the dataset. NOTE: Clients should use the "since"-parameter whenever \
possible.""",
        default=0)
    get_dataset_entities_parser.add_argument(
        "--history",
        default=True,
        help="""If this is true (the default) all entities, including replaced ones, will be returned. \
    If this is false, only the latest version of the entities will be returned""")
    get_dataset_entities_parser.add_argument(
        "--deleted",
        default=True,
        help="""If this is True (the default) deleted entities will also be returned. \
                If this is False, only the non-deleted entities will be returned.""")
    get_dataset_entities_parser.set_defaults(subparser_func=get_dataset_entities_cmd)

    get_dataset_entity_parser = add_subparser(
        "get-dataset-entity",
        description="Returns the specified entity of the specified dataset")
    get_dataset_entity_parser.add_argument(
        "dataset_id",
        help="The id of the dataset.")
    get_dataset_entity_parser.add_argument(
        "entity_id",
        help="The id of the entity.")
    get_dataset_entity_parser.set_defaults(subparser_func=get_dataset_entity_cmd)

    get_pump_parser = add_subparser(
        "get-pump",
        description="Display detailed information about the pump in the specified pipe")
    get_pump_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    get_pump_parser.set_defaults(subparser_func=get_pump_cmd)

    stop_pump_parser = add_subparser(
        "stop-pump",
        description="Stop the pump in the specified pipe")
    stop_pump_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    stop_pump_parser.set_defaults(subparser_func=stop_pump_cmd)

    start_pump_parser = add_subparser(
        "start-pump",
        description="Run the pump in the speficied pipe")
    start_pump_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    start_pump_parser.add_argument(
        "--allow-already-running-pump",
        default=False,
        action='store_true',
        help="This option will make the command ignore an already running pump. Normally, that is treated as an error.")
    start_pump_parser.add_argument(
        "--dont-wait-for-pump-to-start",
        default=False,
        action='store_true',
        help="This option will prevent the command from waiting for the pump to start before returning. This is useful "
             "for batch-operations where you want to schedule lots of pumps to run, but don't wait for the server to "
             "actually start them all.")
    start_pump_parser.add_argument(
        "--wait-timeout",
        dest="wait_timeout",
        default=0,
        type=int,
        help="If this is not zero (the default), the command will wait for the pump to finish running. "
             "If the pump hasn't stopped within the specified number of seconds, an error is raised.")
    start_pump_parser.set_defaults(subparser_func=start_pump_cmd)

    disable_pump_parser = add_subparser(
        "disable-pump",
        description="Disable the pump in the specified pipe")
    disable_pump_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    disable_pump_parser.set_defaults(subparser_func=disable_pump_cmd)

    enable_pump_parser = add_subparser(
        "enable-pump",
        description="Enable the pump in the specified pipe")
    enable_pump_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    enable_pump_parser.set_defaults(subparser_func=enable_pump_cmd)

    wait_for_pump_to_finish_running_parser = add_subparser(
        "wait-for-pump-to-finish-running",
        description="Wait for the pump in the specified pipe to finish running")
    wait_for_pump_to_finish_running_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    wait_for_pump_to_finish_running_parser.add_argument(
        "timeout",
        default=60,
        type=int,
        help="The maximum number of seconds to wait.")
    wait_for_pump_to_finish_running_parser.set_defaults(subparser_func=wait_for_pump_to_finish_running_cmd)

    unset_last_seen_parser = add_subparser(
        "unset-last-seen",
        description="Unset 'last seen' token on the specified pipe")
    unset_last_seen_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    unset_last_seen_parser.set_defaults(subparser_func=unset_last_seen_cmd)

    update_last_seen_parser = add_subparser(
        "update-last-seen",
        description="Update the 'last seen' token on the specified pipe")
    update_last_seen_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    update_last_seen_parser.add_argument(
        "last_seen",
        help="The new last change token. For sdshare-clients, this should be on the form: '2015-07-20T10:00:15Z'")
    update_last_seen_parser.set_defaults(subparser_func=update_last_seen_cmd)

    run_pump_operation_parser = add_subparser(
        "run-pump-operation",
        description="""This is a special generic command that runs the specified operation on the specified pipe.
The usecase is pipes that support operations that aren't part of the 'official'sesam-api, and operations that hasn't
yet gotten their own dedicated command in the command-line client.""")
    run_pump_operation_parser.add_argument(
        "pipe_id",
        help="The id of the pipe.")
    run_pump_operation_parser.add_argument(
        "operation",
        help="The operation to run.")
    run_pump_operation_parser.add_argument(
        "-d",
        action="append",
        metavar="<parametername>=<parametervalue>",
        dest="operation_parameters",
        help='A <parametername>=<parametervalue> pair. This option can be specified zero or more times. '
             'Example: -d last-seen="2015-07-20T10:00:15Z".')
    run_pump_operation_parser.set_defaults(subparser_func=run_pump_operation_cmd)

    export_parser = add_subparser(
        "export",
        description='Exports the complete configuration to a file called "sesam.conf.json" in the given "exportdir" '
                    'directory. If no directory is given then it will be written to the current directory.')
    export_parser.add_argument(
        "exportdir",
        default=".",
        nargs="?",
        help="The directory to put the configuration file(s) in. Defaults to the current directory.")
    export_parser.add_argument(
        "--export-filename",
        default=None,
        help="The filename to use when storing the configuration. Defaults to 'sesam.conf.json'. This option is mutually exclusive with the '--expand' option.")
    export_parser.add_argument(
        "--expand",
        default=False,
        action="store_true",
        help="Instead of writing the configuration to a JSON file, write it to multiple files.  This option is mutually exclusive with the '--export_filename' option.")
    export_parser.add_argument(
        "--force",
        default=False,
        action="store_true",
        help="""Force a "yes" response to all interactive questions.""")
    export_parser.add_argument(
        "--dry-run",
        default=False,
        action="store_true",
        help="""Show which files would be overwritten, deleted or created if the export had run.""")
    export_parser.set_defaults(subparser_func=export_cmd)

    import_parser = add_subparser(
        "import",
        description='Imports the complete configuration from one or more files and or directories containing '
                    '"*.conf.json" files.')
    import_parser.add_argument(
        "config_file_or_folder",
        metavar="config-file-or-folder",
        nargs="+",
        help='One or more configfiles and/or directories containing one or more *.conf.json files.')
    import_parser.add_argument(
        "--dry-run",
        default=False,
        action="store_true",
        help="Show which files would be used if the import had been run, and which errors (if any) would have "
             "resulted.")
    add_force_config_argument(import_parser)
    import_parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help='If specified then the full details of any errors will be displayed. '
             '(By default only a short summary of the errors is shown)')
    import_parser.set_defaults(subparser_func=import_cmd)

    # License stuff

    get_license_parser = add_subparser(
        "get-license",
        description="Returns the current license info for the node")
    get_license_parser.set_defaults(subparser_func=get_license_cmd)

    put_license_parser = add_subparser(
        "put-license",
        description="Replaces the license for the node with the specified license key")
    put_license_parser.add_argument(
        "license_file",
        metavar="path-to-license-key",
        help='The file containing the license key.')
    put_license_parser.set_defaults(subparser_func=put_license_cmd)

    # Secrets manager

    # Get all secrets
    get_secrets_parser = add_subparser(
        "get-secrets",
        description="Returns the current secret for the node")
    get_secrets_parser.add_argument(
        "system_id",
        metavar="system-id",
        nargs="?",
        default=None,
        help='The id of the System to get the secrets from. If this parameter is omitted, the global secrets will '
             'returned.')
    get_secrets_parser.set_defaults(subparser_func=get_secrets_cmd)

    # Puts secrets file
    put_secrets_parser = add_subparser(
        "put-secrets",
        description="Updates the node secret manager with the contents of a key:secret JSON file")
    put_secrets_parser.add_argument(
        "system_id",
        metavar="system-id",
        nargs="?",
        default=None,
        help='The id of the System the secret(s) should apply to. If this parameter is omitted, the secret(s) will '
             'apply to all Systems.')
    put_secrets_parser.add_argument(
        "secrets_file",
        metavar="path-to-secrets-file",
        help='The file containing the secrets in JSON format.')
    put_secrets_parser.add_argument(
        "--dont_encrypt",
        default=False,
        action='store_true',
        help="Don't encrypt the secrets (for testing purposes).")
    put_secrets_parser.set_defaults(subparser_func=put_secrets_cmd)


    # Posts secrets file
    post_secrets_parser = add_subparser(
        "post-secrets",
        description="Updates the node secret manager with the contents of a key:secret JSON file")
    post_secrets_parser.add_argument(
        "system_id",
        metavar="system-id",
        nargs="?",
        default=None,
        help='The id of the System the secret(s) should apply to. If this parameter is omitted, the secret(s) will '
             'apply to all Systems.')
    post_secrets_parser.add_argument(
        "secrets_file",
        metavar="path-to-secrets-file",
        help='The file containing the secrets in JSON format.')
    post_secrets_parser.add_argument(
        "--dont_encrypt",
        default=False,
        action='store_true',
        help="Don't encrypt the secrets (for testing purposes).")
    post_secrets_parser.set_defaults(subparser_func=post_secrets_cmd)


    # Delete a secret
    delete_secret_parser = add_subparser(
        "delete-secret",
        description="Deletes the given secret from the node")
    delete_secret_parser.add_argument(
        "system_id",
        metavar="system-id",
        nargs="?",
        default=None,
        help='The id of the System the secret should deleted from. If this parameter is omitted, the secret will '
             'be deleted from the global secrets.')
    delete_secret_parser.add_argument(
        "secret_key",
        metavar="secret-key",
        help='The secret key to delete.')
    delete_secret_parser.set_defaults(subparser_func=delete_secret_cmd)

    # Env vars manager

    # Get all env vars
    get_env_vars_parser = add_subparser(
        "get-env-vars",
        description="Returns the current environment variables for the node")
    get_env_vars_parser.set_defaults(subparser_func=get_env_vars_cmd)

    # Put env vars file
    put_env_vars_parser = add_subparser(
        "put-env-vars",
        description="Replaces the contents of the node environment variable manager with the contents of a "
                    "environment-var:value JSON file")
    put_env_vars_parser.add_argument(
        "env_vars_file",
        metavar="path-to-env-vars-file",
        help='The file containing the environment variables in JSON format.')
    put_env_vars_parser.set_defaults(subparser_func=put_env_vars_cmd)

    # Post env vars file
    post_env_vars_parser = add_subparser(
        "post-env-vars",
        description="Updates the node environment variable manager with the contents of a "
                    "environment-var:value JSON file")
    post_env_vars_parser.add_argument(
        "env_vars_file",
        metavar="path-to-env-vars-file",
        help='The file containing the environment variables in JSON format.')
    post_env_vars_parser.set_defaults(subparser_func=post_env_vars_cmd)

    # Delete a env var
    delete_env_var_parser = add_subparser(
        "delete-env-var",
        description="Deletes the given environment variable from the node")
    delete_env_var_parser.add_argument(
        "env_var",
        metavar="env-var",
        help='The environment variable to delete.')
    delete_env_var_parser.set_defaults(subparser_func=delete_env_var_cmd)

    # log in to a sesam installation
    login_parser = add_subparser(
        "login",
        description="Log in to a sesam installation",
        add_default_output_format_argument=False)
    login_parser.set_defaults(subparser_func=login_cmd)

    # log out from a sesam installation
    logout_parser = add_subparser(
        "logout",
        description="Log out from a sesam installation",
        add_default_output_format_argument=False)
    logout_parser.set_defaults(subparser_func=logout_cmd)

    return parser


class BodyDataExtractingHTMLParser(HTMLParser):
    def error(self, message):
        self.parser_error_messages.append(message)

    def __init__(self):
        super().__init__()
        self.parser_error_messages = []
        self.is_inside_body_tag = False
        self.body_data = []

    def handle_starttag(self, tag, attrs):
        if tag == "body":
            self.is_inside_body_tag = True

    def handle_endtag(self, tag):
        if tag == "body":
            self.is_inside_body_tag = False

    def handle_data(self, data):
        if self.is_inside_body_tag:
            data = data.strip()
            if data:
                if not data.endswith("."):
                    data += "."
                self.body_data.append(data)


def main():
    parser = get_parser()
    args = parser.parse_args()

    loglevel = {
        'ERROR': logging.ERROR,
        'WARNING': logging.WARNING,
        'INFO': logging.INFO,
        'DEBUG': logging.DEBUG}[args.loglevel]

    logging.basicConfig(level=loglevel,
                        format="%(asctime)s %(levelname)-5.5s [%(name)s][%(threadName)s] %(message)s")

    logger.info("main() running")
    configfilename = get_configfilename()
    logger.info("Using config file '%s'" % (configfilename,))

    myconfigparser = get_configparser()
    sesamapi_base_url = myconfigparser.get("server",
                                           "base_url",
                                           fallback=getattr(args, "server_base_url", ""))

    logger.info("sesamapi_base_url: '%s'" % (sesamapi_base_url,))

    if hasattr(args, "subparser_func"):
        # A command was specified

        # check if the subparser_func wants a Connection parameter.
        argspec = inspect.getfullargspec(args.subparser_func)
        if argspec.args[0] == "connection":
            # Ok, this function wants a connection argument
            sesamapi_base_url = getattr(args, "server_base_url")
            if not sesamapi_base_url:
                sesamapi_base_url = myconfigparser.get("server", "base_url", fallback="")
            if not sesamapi_base_url:
                main_utils.print_message(
                    "No server base_url specified! Either use the 'sesam config server.base_url "
                    "<baseurl>' command to specify a default base_url, "
                    "or specify the base_url with the '--server_base_url' parameter. "
                    "Falling back to default value ('http://localhost:9042/api')",
                    file=sys.stderr)
                sesamapi_base_url = "http://localhost:9042/api"

            username = getattr(args, "server_username", myconfigparser.get("server", "username", fallback=""))
            password = getattr(args, "server_password", myconfigparser.get("server", "password", fallback=""))

            jwt_auth_token = None
            jwt_filename = args.jwt
            if jwt_filename:
                if not os.path.exists(jwt_filename):
                    exit_with_error("The file '%s' specified with the '--jwt' argument doesn't exist!" % (
                        jwt_filename,))

                try:
                    with open(jwt_filename, "r") as jwt_file:
                        jwt_auth_token = jwt_file.read().strip()

                except IOError as e:
                    exit_with_error("Failed to open the file '%s' specified with the '--jwt' argument: %s" % (
                        jwt_filename, e))

            if jwt_auth_token is None:
                # No jwt-file specified as an commandline argument, so check if we have one stored in the
                # configfile (as a result of an earlier 'sesam login' command).
                jwt_auth_token = myconfigparser.get("server", "jwt", fallback=None)
                got_jwt_from_commandline_argument = False
            else:
                got_jwt_from_commandline_argument = True

            if args.command in ("login", "logout"):
                # This command is special and does not use a jwt token.
                jwt_auth_token = None

            timeout = int(myconfigparser.get("server", "timeout", fallback="") or 30)
            connection = Connection(sesamapi_base_url=sesamapi_base_url,
                                    username=username,
                                    password=password,
                                    timeout=timeout,
                                    jwt_auth_token=jwt_auth_token
                                    )
            try:
                args.subparser_func(connection, args)

            except PumpDoesNotSupportTheOperation as e:
                # This happens if the user user has tried to perform some operation ("start", "stop", "enable",
                # "disable", "update-last-seen", etc) on a pump, but the pump doesn't support that operation.
                # Rather than have each command-function handle this exception on its own, we catch it here,
                # so that we only need to handle it in one common place.
                supported_operations = ["'%s'" % (operation, ) for operation in e.pump.supported_operations]
                if len(supported_operations) == 0:
                    supported_operations_msg = "It doesn't support any operations."
                elif len(supported_operations) == 1:
                    supported_operations_msg = "It only supports this operation: '%s'." % (supported_operations[0],)
                else:
                    two_last_operations_string = " and ".join(supported_operations[-2:])
                    all_operations_string = ", ".join(supported_operations[:-2] + [two_last_operations_string])
                    supported_operations_msg = "It only supports these operations: %s." % (all_operations_string,)

                exit_with_error(
                    "The pump doesn't support the '%s' operation. %s" % (
                        e.operation,
                        supported_operations_msg
                    ))

            except requests.exceptions.ConnectionError as e:
                exit_with_error(
                    "Failed to connect to the sesam installation at '%s'! The following error occurred: %s" % (
                        sesamapi_base_url, e))

            except BadRequestException as e:
                if e.response.status_code in (401, 403):
                    # The user is not authenticated or isn't allowed to perform some operation.

                    # The response from the server looks something like this:
                    #
                    #  <html>
                    #   <head>
                    #    <title>401 Unauthorized</title>
                    #   </head>
                    #   <body>
                    #    <h1>401 Unauthorized</h1>
                    #    This server could not verify that you are authorized to access the document you requested.  Either you supplied the wrong credentials (e.g., bad password), or your browser does not understand how to supply the credentials required.<br/><br/>
                    #  The user does not have the &#x27;Read configuration&#x27; permission.
                    #   </body>
                    #  </html>
                    #
                    # or like this:
                    #
                    #   <html>
                    #   <head>
                    #    <title>403 Forbidden</title>
                    #   </head>
                    #   <body>
                    #    <h1>403 Forbidden</h1>
                    #    Access was denied to this resource.<br/><br/>
                    #  The user does not have the &#x27;Read configuration&#x27; permission.
                    #   </body>
                    #  </html>
                    #
                    #
                    # or like this:
                    #
                    # {
                    #   status: 401,
                    #   type: "about:blank",
                    #   title: "Unauthorized",
                    #   detail: "The user does not have the 'Read configuration' permission."
                    # }
                    # We want to extract the text from the body-tag and display just that to the user.

                    from html.parser import HTMLParser
                    parser = BodyDataExtractingHTMLParser()
                    response_text = e.response.text
                    parser.feed(response_text)
                    error_message = " ".join(parser.body_data)
                    if parser.parser_error_messages or not error_message:
                        # This is not html, lets try parsing it as json
                        try:
                            response_json = e.response.json()
                            error_message = "Got a '%s %s' response: %s" % (response_json["status"], response_json["title"], response_json["detail"])
                        except:
                            # Failed to parse the reply message as either html or json. hmm. Lets just use the raw response text.
                            error_message = response_text
                    exit_with_error(error_message)

                # This is some other error
                raise

        else:
            # This function does not want a connection argument
            args.subparser_func(args)

    else:
        # no command was specified
        parser.print_usage()

if __name__ == "__main__":
    if sys.version_info[0] < 3:
        print("Sesam client only supports Python 3.x")
        sys.exit(1)

    main()



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
from ..utils import print_item, exit_with_error, print_message

from ...exceptions import PumpIsAlreadyRunning


def get_pipe_and_exit_if_missing(connection, args):
    pipe = connection.get_pipe(args.pipe_id)
    if pipe is None:
        exit_with_error("No pipe with id '%s' was found!" % (args.pipe_id,))
    return pipe


def get_pump_cmd(connection, args):
    pipe = get_pipe_and_exit_if_missing(connection, args)
    pump = pipe.get_pump()
    print_item(pump, args)


def start_pump_cmd(connection, args):
    pipe = get_pipe_and_exit_if_missing(connection, args)

    pump = pipe.get_pump()
    try:
        pump.start(dont_wait_for_pump_to_start=args.dont_wait_for_pump_to_start)
    except PumpIsAlreadyRunning:
        if not args.allow_already_running_pump:
            exit_with_error("The pump of the pipe id '%s' is already running!" % (args.pipe_id,))

    if args.wait_timeout > 0:
        pump.wait_for_pump_to_finish_running(timeout=args.wait_timeout)

    if args.out in ("json", "yaml"):
        print_item(pump, args)

    elif args.out == "plain":
        print_message("Started the pump in this pipe:")
        print_message(str(pipe))
        print_message("=>", indent=2)
        print_message(str(pump), indent=4)

    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))


def stop_pump_cmd(connection, args):
    pipe = get_pipe_and_exit_if_missing(connection, args)

    pump = pipe.get_pump()

    pump.stop()

    if args.out in ("json", "yaml"):
        print_item(pump, args)

    elif args.out == "plain":
        print_message("Stopped this  pump:")
        print_message(str(pump))

    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))


def disable_pump_cmd(connection, args):
    pipe = get_pipe_and_exit_if_missing(connection, args)

    pump = pipe.get_pump()
    pump.disable()

    if args.out in ("json", "yaml"):
        print_item(pump, args)

    elif args.out == "plain":
        print_message("Disabled this pump:")
        print_message(str(pump))

    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))


def enable_pump_cmd(connection, args):
    pipe = get_pipe_and_exit_if_missing(connection, args)

    pump = pipe.get_pump()
    pump.enable()
    if args.out in ("json", "yaml"):
        print_item(pump, args)

    elif args.out == "plain":
        print_message("Enabled this pipe:")
        print_message(str(pipe))

    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))


def wait_for_pump_to_finish_running_cmd(connection, args):
    pipe = get_pipe_and_exit_if_missing(connection, args)
    pump = pipe.get_pump()
    pump.wait_for_pump_to_finish_running(timeout=args.timeout)

    if args.out in ("json", "yaml"):
        print_item(pump, args)

    elif args.out == "plain":
        print_message("Waited for this pipe:")
        print_message(str(pipe))

    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))


def unset_last_seen_cmd(connection, args):
    pipe = get_pipe_and_exit_if_missing(connection, args)

    pump = pipe.get_pump()
    pump.unset_last_seen()
    if args.out in ("json", "yaml"):
        print_item(pump, args)
    elif args.out == "plain":
        print_message("Updated this pump:")
        print_message(str(pump))
    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))


def update_last_seen_cmd(connection, args):
    pipe = get_pipe_and_exit_if_missing(connection, args)

    pump = pipe.get_pump()
    pump.update_last_seen(args.last_seen)
    if args.out in ("json", "yaml"):
        print_message(pump.as_json())
    elif args.out == "plain":
        print_message("Updated this pump:")
        print_message(str(pump))
    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))


def run_pump_operation_cmd(connection, args):
    pipe = get_pipe_and_exit_if_missing(connection, args)

    pump = pipe.get_pump()
    operation_parameters_list = args.operation_parameters
    operation_parameters = {}

    def strip_and_remove_enclosing_quotes(value):
        value = value.strip()
        if (value.startswith("'") and value.endswith("'")) or (value.startswith('"') and value.endswith('"')):
            value = value[1:-1]
            value = value.strip()
        return value

    if operation_parameters_list:
        for operation_parameter in operation_parameters_list:
            parameter_name, parameter_value = operation_parameter.split("=", 1)
            parameter_name = strip_and_remove_enclosing_quotes(parameter_name)
            parameter_value = strip_and_remove_enclosing_quotes(parameter_value)
            operation_parameters[parameter_name] = parameter_value

    pump.run_operation(args.operation, operation_parameters)
    if args.out in ("json", "yaml"):
        print_message(pump.as_json())
    elif args.out == "plain":
        print_message("Updated this pump:")
        print_message(str(pump))
    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import pprint

from ..utils import print_item, print_item_list, exit_with_error
from ...exceptions import PipeAlreadyExists


def get_pipes_cmd(connection, args):
    # TODO: filtering
    pipes = connection.get_pipes()
    print_item_list(pipes, args)


def get_pipe_cmd(connection, args):
    pipe = connection.get_pipe(args.pipe_id)
    if pipe is None:
        exit_with_error("No pipe with id '%s' was found!" % (args.pipe_id,))
    print_item(pipe, args)


def get_pipe_schema_def_cmd(connection, args):
    sample_size = args.sample_size
    keys_only = args.keys_only

    pipe = connection.get_pipe(args.pipe_id)
    if pipe is None:
        exit_with_error("No pipe with id '%s' was found!" % (args.pipe_id,))

    schema_def = connection.get_pipe_schema_def(args.pipe_id, sample_size, keys_only)
    print_item_list(schema_def, args)


def add_pipes_cmd(connection, args):
    try:
        pipes = connection.add_pipes(args.pipe_configurations, force=args.force)
        print_item_list(pipes, args)
    except PipeAlreadyExists as e:
        exit_with_error("One or more of the pipes already exist: %s" % (e,))


def modify_pipe_cmd(connection, args):
    pipe_config = args.pipe_configuration
    pipe_id = pipe_config.get("_id")
    if not pipe_id:
        exit_with_error("The pipe-configuration doesn't contain a '_id' attribute! pipe_config: \n%s" % (
            pprint.pformat(pipe_config),))

    pipe = connection.get_pipe(pipe_id)
    if pipe is None:
        exit_with_error("No pipe with id '%s' was found!" % (pipe_id,))

    pipe.modify(pipe_config, force=args.force)
    print_item(pipe, args)


def delete_pipe_cmd(connection, args):
    pipe = connection.get_pipe(args.pipe_id)
    if pipe is None:
        exit_with_error("No pipe with id '%s' was found!" % (args.pipe_id,))
    pipe.delete()
    print_item(pipe, args)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
from .connection import Connection
from .pipe import Pipe
from .dataset import Dataset
from .exceptions import PumpIsAlreadyRunning


# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import getpass
import os

from .config import get_configfilename, get_configparser, save_configparser_to_file
from ..utils import print_message

_jwt_configfile_section = "server"
_jwt_configfile_option = "jwt"


def logout_cmd(connection, args):
    myconfigparser = get_configparser()
    current_jwt = myconfigparser.get(_jwt_configfile_section, _jwt_configfile_option, fallback=None)
    if current_jwt is None:
        print_message("The sesam commandline client is not currently logged in.")
    else:
        myconfigparser.remove_option(_jwt_configfile_section, _jwt_configfile_option)
        save_configparser_to_file(myconfigparser)
        print_message("The sesam commandline client is now logged out.")


def login_cmd(connection, args):
    print_message("Login with the email address and password you used to register an account in the Sesam Portal. Warning: this will store the resulting authorization token in the file '%s'." % (
        get_configfilename()
    ))
    myconfigparser = get_configparser()

    email = input("Email: ")
    password = getpass.getpass(prompt="Password:")

    jwt_auth_token = connection.login(email, password)

    if not myconfigparser.has_section(_jwt_configfile_section):
        myconfigparser.add_section(_jwt_configfile_section)

    myconfigparser.set(_jwt_configfile_section, _jwt_configfile_option, jwt_auth_token)
    save_configparser_to_file(myconfigparser)

    print_message(
        "Logged in successfully and stored the new authorization token in the file '%s'." % (
            get_configfilename()
        ))


# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import os

from ..utils import print_item, exit_with_error


def get_license_cmd(connection, args):
    license_info = connection.get_license()
    print_item(license_info, args)


def put_license_cmd(connection, args):
    license_path = args.license_file
    if not os.path.exists(license_path):
        exit_with_error("The specified path '%s' doesn't exist!" % (license_path,))
    with open(license_path) as licence_content:
        license_info = connection.set_license(licence_content)
        print_item(license_info, args)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.


import collections
import io
import json
import os
import zipfile

from .. import utils


def export_cmd(connection, args):
    original_exportdir = args.exportdir or "."

    exportdir = os.path.abspath(original_exportdir)

    files_that_will_be_deleted = []  # filenames
    files_that_will_be_added = collections.OrderedDict()  # filename => filecontent
    files_that_will_be_modified = collections.OrderedDict()   # filename => new filecontent
    os.makedirs(exportdir, exist_ok=True)

    source_files = collections.OrderedDict()  # filename => filecontent
    if args.expand:
        if args.export_filename:
            utils.exit_with_error("The '--expand' and '--export_filename' are mutually exclusive! They can't both be specified.")

        for dirpath, dirnames, filenames in os.walk(exportdir):
            for filename in filenames:
                if not utils.is_valid_configfile_name(filename):
                    continue
                files_that_will_be_deleted.append(os.path.abspath(os.path.join(dirpath, filename)))

        config_as_zip = connection.get_config_as_zip()
        config_zipfile = zipfile.ZipFile(io.BytesIO(config_as_zip))
        for filename in config_zipfile.namelist():
            source_file_content = config_zipfile.read(filename)
            source_files[filename] = source_file_content
    else:
        # export everything together as one file
        source_filename = args.export_filename or "sesam.conf.json"
        assert utils.is_valid_configfile_name(source_filename)
        source_files[source_filename] = connection.get_config_as_json_string().encode("utf-8")

        target_filename = os.path.abspath(os.path.join(exportdir, source_filename))
        if os.path.exists(target_filename):
            files_that_will_be_deleted.append(target_filename)

    for source_filename, source_file_content in source_files.items():
        target_filename = os.path.abspath(os.path.join(exportdir, source_filename))
        if target_filename in files_that_will_be_deleted:
            files_that_will_be_deleted.remove(target_filename)
            # check if the file will be modified
            with open(target_filename, "rb") as target_file:
                target_file_content = target_file.read()
                if target_file_content != source_file_content:
                    # the raw file-content has been changed, but if both files are valid json and the parsed
                    # content is identical we ignore the change, since the files are functionally identical.
                    try:
                        source_file_content_parsed = json.loads(source_file_content.decode("utf-8"))
                        target_file_content_parsed = json.loads(target_file_content.decode("utf-8"))
                        is_changed = target_file_content_parsed != source_file_content_parsed
                    except:
                        # failed to parse at least one of the files, so we will assume the files are functionally
                        # different.
                        is_changed = True
                    if is_changed:
                        files_that_will_be_modified[target_filename] = source_file_content
        else:
            files_that_will_be_added[target_filename] = source_file_content

    def get_filenames_as_humanreadable_list(filenames):
        result = ""
        for filename in filenames:
            if not os.path.isabs(original_exportdir):
                # The user specified a relative path, so use a relative path when printing the message, too.
                filename = os.path.relpath(filename, start=os.getcwd())
            result += "  " + filename + "\n"
        return result

    def get_filelist_output(verb):
        result = ""
        if files_that_will_be_deleted:
            result += "The following files %s deleted:\n" % (verb,)
            result += get_filenames_as_humanreadable_list(files_that_will_be_deleted)

        if files_that_will_be_modified:
            result += "The following files %s modified:\n" % (verb,)
            result += get_filenames_as_humanreadable_list(files_that_will_be_modified)

        if files_that_will_be_added:
            result += "The following files %s added:\n" % (verb,)
            result += get_filenames_as_humanreadable_list(files_that_will_be_added)
        return result

    if files_that_will_be_deleted or files_that_will_be_modified or files_that_will_be_added:
        if args.dry_run:
            utils.print_message("The '--dry-run' parameter was specified, so no files will be touched. "
                                "Otherwise, the following changes would have been made:")
            utils.print_message(get_filelist_output("would have been"))
        else:
            if not args.force:
                prompt = get_filelist_output("will be")

                prompt += "Are you sure you want to do this (yes/no)"

                answer = input(prompt)
                if answer.lower() not in ("y", "yes"):
                    utils.print_message("Exiting without making any changes.")
                    return

            for filename, filecontent in files_that_will_be_added.items():
                os.makedirs(os.path.dirname(filename), exist_ok=True)
                with open(filename, "wb") as targetfile:
                    targetfile.write(filecontent)

            for filename, filecontent in files_that_will_be_modified.items():
                with open(filename, "wb") as targetfile:
                    targetfile.write(filecontent)

            for filename in files_that_will_be_deleted:
                os.remove(filename)

            utils.print_message(get_filelist_output("were"))

    else:
        utils.print_message("The existing config-files are up-to-date.")



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.

import io
import json
import os
import pprint
import sys
import zipfile

import collections

from .. import utils
from ..utils import print_message, exit_with_error
from ...exceptions import ConfigUploadFailed


def _get_humanreadable_filename(filename):
    if not os.path.isabs(filename):
        # The user specified a relative path, so use a relative path when printing the message, too.
        filename = os.path.relpath(filename, start=os.getcwd())
    return filename


def _get_filenames_as_humanreadable_list(filenames):
    result = ""
    for filename in filenames:
        filename = _get_humanreadable_filename(filename)
        result += "  " + filename + "\n"
    return result


def import_cmd(connection, args):
    for config_file_or_folder in args.config_file_or_folder:
        if not os.path.exists(config_file_or_folder):
            exit_with_error("The specified path '%s' doesn't exist!" % (config_file_or_folder,))

    config_files = []
    for config_file_or_folder in args.config_file_or_folder:
        if os.path.isfile(config_file_or_folder):
            config_files.append(config_file_or_folder)
        elif os.path.isdir(config_file_or_folder):
            # Search for all *.conf.json files in the folder
            for dirpath, dirnames, filenames in os.walk(config_file_or_folder):
                for testfilename in filenames:
                    if utils.is_valid_configfile_name(testfilename):
                        if utils.is_valid_but_deprecated_configfile_name(testfilename):
                            print_message(("Warning: The config filename '%s' is deprecated. "
                                           "Configfiles should have names on the form '*.conf.json'.") % (
                                os.path.join(dirpath, testfilename),))
                        testfilename = os.path.join(dirpath, testfilename)
                        config_files.append(testfilename)
        else:
            exit_with_error("The specified path '%s' is neither a file nor a folder!" % (config_file_or_folder,))

    if not config_files:
        exit_with_error("No '*.conf.json' configfiles were found!")

    config_files.sort()

    configentity_id2filename = collections.OrderedDict()

    for config_filename in config_files:
        humanreadable_config_filename = _get_humanreadable_filename(config_filename)
        with open(config_filename, "rt") as config_file:
            config_file_content = None
            try:
                config_file_content = json.load(config_file)
            except Exception as e:
                exit_with_error("Failed to parse the content of the file '%s': %s" % (
                                humanreadable_config_filename, e))

            if not isinstance(config_file_content, list):
                config_file_content = [config_file_content]

            for configentity in config_file_content:
                if "_id" not in configentity:
                    exit_with_error("The file '%s' contains an configentity that doesn't have an '_id' attribute: %s"
                                    % (humanreadable_config_filename, configentity))

                # Check that the ids are unique
                configentity_id = configentity["_id"]
                if configentity_id in configentity_id2filename:
                    otherfilename = configentity_id2filename[configentity_id]
                    if otherfilename == config_filename:
                        exit_with_error("The file '%s' contains two config entities with the id '%s'! "
                                        "The '_id'-values must be unique!"
                                        % (humanreadable_config_filename, configentity_id))
                    else:
                        exit_with_error("The config entity id '%s' was found in both in '%s' and in '%s'! "
                                        "The '_id'-values must be unique!"
                                        % (configentity_id, _get_humanreadable_filename(otherfilename),
                                           humanreadable_config_filename))

                configentity_id2filename[configentity_id] = config_filename

    def print_validation_errors(validation_errors, file=None):
        # validation_errors is a list on this format:
        # [{'config-errors': [{'elements': '$', 'level': 'error',
        #            'msg': "Unknown component id 'longrunningdummytask' of type 'task'"}],
        #      'is-valid-config': False,
        #      'json-schema': {'properties': {'type': {'enum': ['pipe', 'system:sqlite', 'system:sesam-node', ... },
        #      'posted-config': {'_id': 'task:longrunningdummytask1', 'schedule_interval': 60000, ...}}]
        for validation_error in validation_errors:
            parse_error = False
            posted_config = validation_error.get("posted-config")
            if not isinstance(posted_config, dict):
                parse_error = True
                origin_configentity_id = "*unknown*"
                origin_filename = "*unknown*"
            else:
                origin_configentity_id = posted_config.get("_id") or "*unknown*"
                origin_filename = configentity_id2filename.get(origin_configentity_id)
                if origin_filename:
                    origin_filename = _get_humanreadable_filename(origin_filename)
                else:
                    origin_filename = "*unknown*"

            print_message("Issues with the config entity '%s' from the file '%s':" % (origin_configentity_id,
                                                                                      origin_filename, ),
                          file=file,
                          indent=4)
            for error in validation_error["config-errors"]:
                level = error["level"]
                msg = error["msg"]
                elements = error.get("elements")
                elements_term = ""
                if elements:
                    elements_term = " (path: %s)" % (elements,)
                print_message("%s: %s%s" % (level, msg, elements_term),
                              file=file,
                              indent=8)

            if args.verbose or parse_error:
                print_message("Raw validation error:", file=file, indent=6)
                print_message(pprint.pformat(validation_error), file=file, indent=8)

    print_message("Read %d config entities from these config-files:" % (len(configentity_id2filename), ))
    print_message(_get_filenames_as_humanreadable_list(config_files))

    if args.dry_run:
        print_message("\nThe '--dry-run' parameter was specified, so no changes will be done to the configuration. ")
    else:
        if len(config_files) == 1:
            # There is just the one config-file, so just upload that
            config_to_upload = config_files[0]
        else:
            # There is more than one configfile, so upload them all in a ziparchive
            zipped_config = io.BytesIO()
            ziparchive = zipfile.ZipFile(zipped_config, "w")
            for config_filename in config_files:
                humanreadable_config_filename = _get_humanreadable_filename(config_filename)
                ziparchive.write(config_filename, humanreadable_config_filename)
            ziparchive.close()
            config_to_upload = zipped_config

        try:
            response = connection.upload_config(config_to_upload,
                                                force=args.force)
            response_validation_errors = response["validation_errors"]
            if response_validation_errors:
                print_message("The configuration was applied, but these errors/warnings were detected:")
                print_validation_errors(response_validation_errors)

        except ConfigUploadFailed as e:
            response = e.parsed_response
            response_validation_errors = response["validation_errors"]
            if response_validation_errors:
                print_message("Failed to apply the configuration. These errors were detected:", file=sys.stderr)
                print_validation_errors(response_validation_errors, file=sys.stderr)
                exit_with_error(None)



# -*- coding: utf-8 -*-
# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.

import contextlib
import io
import json
import os.path
import pprint
import sys
import tempfile
import zipfile
import types
from unittest import mock

import sesamclient.main.commands.export


@contextlib.contextmanager
def captured_output():
    new_out, new_err = io.StringIO(), io.StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield new_out, new_err
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class MockConnection:
    def get_config_as_zip(self):
        thezip = io.BytesIO()
        thezipfile = zipfile.ZipFile(thezip, "w")
        thezipfile.writestr("metadata.conf.json", json.dumps({
            "_id": "node"
        }))
        thezipfile.writestr("pipes/pipe1.conf.json", json.dumps({
            "_id": "pipe1"
        }))
        thezipfile.writestr("pipes/pipe2.conf.json", json.dumps({
            "_id": "pipe2"
        }))
        thezipfile.writestr("systems/system1.conf.json", json.dumps({
            "_id": "system1"
        }))
        thezipfile.writestr("systems/system2.conf.json", json.dumps({
            "_id": "system2"
        }))
        thezipfile.close()
        return thezip.getvalue()

    def get_config(self):
        return [
            {
                "_id": "node",
                "type": "metadata"
            },
            {
                "_id": "pipe1",
                "type": "pipe"
            },
            {
                "_id": "pipe2",
                "type": "pipe"
            },
            {
                "_id": "system1",
                "type": "system:sqlite"
            },
            {
                "_id": "system2",
                "type": "system:sqlite"
            },
        ]

    def get_config_as_json_string(self):
        return json.dumps(self.get_config(), indent=4, sort_keys=True)


class InputGuard:
    """This class is used to mock.patch() the input() function in cases where we don't expect the client to
    prompt the user for input."""

    def __init__(self, error_msg="The input() method is not supposed to be called in this test!"):
        self._error_msg = error_msg

    def __call__(self, prompt):
        raise AssertionError(self._error_msg)


def test_export_as_single_file_with_user_prompt():
    do_export_as_single_file(force=False)


def test_export_as_single_file_with_force_flag():
    do_export_as_single_file(force=True)


def do_export_as_single_file(force):
    with tempfile.TemporaryDirectory() as exportdir:
        connection = MockConnection()
        args = types.SimpleNamespace()
        args.exportdir = exportdir
        args.export_filename = None
        args.force = force
        args.expand = False
        args.dry_run = False

        if force:
            # The force-flag is set, so the client shouldn't prompt the user for anything
            input_mock = InputGuard()
        else:
            input_mock = mock.Mock(return_value="yes")

        with captured_output() as (out, err):
            with mock.patch("builtins.input", new=input_mock):
                sesamclient.main.commands.export.export_cmd(connection, args)

                if not force:
                    # Check that the user was asked for a confirmation
                    assert input_mock.call_count == 1, "The client should have asked the user to confirm the modifications to the files!"

                filename = "sesam.conf.json"
                assert os.path.exists(os.path.join(exportdir, filename)), "The file '%s' should have been created by the export!" % (filename,)

            out = out.getvalue()
            err = err.getvalue()

            assert "error" not in out
            assert "error" not in err

            assert """\
The following files were added:
  %s/sesam.conf.json""" % (exportdir,) in out
            assert "The following files were modified" not in out
            assert "The following files were deleted" not in out
            assert "The following files were modified" not in err
            assert "The following files were deleted" not in err

        ##################################################################################
        # check that a second run doesn't result in any modified, deleted or added files
        ##################################################################################
        with captured_output() as (out, err):
            with mock.patch("builtins.input",
                            new=InputGuard("The input() method is not supposed to be called in this test, since no "
                                           "changes to the files is supposed to happen in the second export!")):
                args.force = False
                sesamclient.main.commands.export.export_cmd(connection, args)
                filename = "sesam.conf.json"
                assert os.path.exists(os.path.join(exportdir, filename)), "The file '%s' should still exist!" % (filename,)

            out = out.getvalue()
            err = err.getvalue()

            assert "error" not in out
            assert "error" not in err

            assert "The existing config-files are up-to-date" in out
            assert "The following files were added" not in out
            assert "The following files were added" not in err
            assert "The following files were modified" not in out
            assert "The following files were deleted" not in out
            assert "The following files were modified" not in err
            assert "The following files were deleted" not in err


def test_export_as_multiple_files_with_user_prompt():
    do_export_as_multiple_files(force=False)


def test_export_as_multiple_files_with_force_flag():
    do_export_as_multiple_files(force=True)


def do_export_as_multiple_files(force):
    with tempfile.TemporaryDirectory() as exportdir:
        connection = MockConnection()
        args = types.SimpleNamespace()
        args.exportdir = exportdir
        args.force = force
        args.expand = True
        args.dry_run = False
        args.export_filename = None

        test_existing_non_config_file_that_should_not_be_touched_filename = os.path.join(exportdir, "test_existing_non_config_file_that_should_not_be_touched.json")
        with open(test_existing_non_config_file_that_should_not_be_touched_filename, "w") as testfile:
            testfile.write("This file should not be touched by the export")
        assert os.path.exists(test_existing_non_config_file_that_should_not_be_touched_filename)

        test_existing_file_that_should_be_deleted_filename = os.path.join(exportdir, "testfilethatshouldbedeleted.conf.json")
        with open(test_existing_file_that_should_be_deleted_filename, "w") as testfile:
            testfile.write("This file should be deleted by the export")
        assert os.path.exists(test_existing_file_that_should_be_deleted_filename)

        test_existing_file_that_should_be_modified_filename = os.path.join(exportdir, "pipes/pipe1.conf.json")
        os.makedirs(os.path.dirname(test_existing_file_that_should_be_modified_filename), exist_ok=True)
        with open(test_existing_file_that_should_be_modified_filename, "w") as testfile:
            testfile.write("This file should be modified by the export")
        assert os.path.exists(test_existing_file_that_should_be_modified_filename)

        if force:
            # The force-flag is set, so the client shouldn't prompt the user for anything
            input_mock = InputGuard
        else:
            input_mock = mock.Mock(return_value="yes")

        with captured_output() as (out, err):
            with mock.patch("builtins.input", new=input_mock):
                sesamclient.main.commands.export.export_cmd(connection, args)

            if not force:
                # Check that the user was asked for a confirmation
                assert input_mock.call_count == 1, "The client should have asked the user to confirm the modifications to the files!"

            existing_files = []
            for dirpath, dirnames, filenames in os.walk(exportdir):
                for filename in filenames:
                    existing_files.append(os.path.relpath(os.path.join(dirpath, filename), start=exportdir))

            for filename in ["pipes/pipe1.conf.json",
                             "pipes/pipe2.conf.json",
                             "systems/system1.conf.json",
                             "systems/system1.conf.json",
                             "metadata.conf.json"]:
                assert os.path.exists(os.path.join(exportdir, filename)), "The file '%s' should have been created by the export! existing_files:\n%s" % (
                                filename,
                                pprint.pformat(existing_files))

            assert not os.path.exists(test_existing_file_that_should_be_deleted_filename), "The file '%s' should have been deleted by the export!" % (
                             test_existing_file_that_should_be_deleted_filename)

            out = out.getvalue()
            err = err.getvalue()

            assert "error" not in out
            assert "error" not in err

            assert """\
The following files were added:
  %(exportdir)s/metadata.conf.json
  %(exportdir)s/pipes/pipe2.conf.json
  %(exportdir)s/systems/system1.conf.json
  %(exportdir)s/systems/system2.conf.json
""" % {"exportdir": exportdir} in out
            assert ("The following files were deleted:\n  %(exportdir)s/testfilethatshouldbedeleted.conf.json" %
                          {"exportdir": exportdir} in out)
            assert ("The following files were modified:\n  %(exportdir)s/pipes/pipe1.conf.json" %
                          {"exportdir": exportdir} in out)

        # check that the unrelated json-file isn't touched.
        assert os.path.exists(test_existing_non_config_file_that_should_not_be_touched_filename)

        ##################################################################################
        # check that a second run doesn't result in any modified, deleted or added files
        ##################################################################################
        with captured_output() as (out, err):
            with mock.patch("builtins.input",
                            new=InputGuard("The input() method is not supposed to be called in this test, since no "
                                           "changes to the files is supposed to happen in the second export!")):
                args.force = False
                sesamclient.main.commands.export.export_cmd(connection, args)

            out = out.getvalue()
            err = err.getvalue()

            assert "error" not in out
            assert "error" not in err

            assert "The existing config-files are up-to-date" in out
            assert "The following files were added" not in out
            assert "The following files were added" not in err
            assert "The following files were modified" not in out
            assert "The following files were deleted" not in out
            assert "The following files were modified" not in err
            assert "The following files were deleted" not in err



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import io
import json
import sys
import time
import urllib.parse
import zipfile
import threading
import socket

import requests
import requests.adapters
from urllib3.connection import HTTPConnection

from oauthlib.oauth2 import BackendApplicationClient, TokenExpiredError
from requests_oauthlib import OAuth2Session

from . import utils, entity_json
from .dataset import Dataset
from .exceptions import PipeAlreadyExists, SystemAlreadyExists, ConfigUploadFailed, \
    TimeoutWhileWaitingForBackupToFinishException, ServiceUnavailable, InternalServerError, BadRequestException, \
    TimeoutWhileWaitingForRestoreToFinishException
from .log import Log
from .pipe import Pipe
from .system import System
from .model import Model
from .utils import validate_response_is_ok
from requests.exceptions import RequestException


def quote_id_for_url(item_id):
    return urllib.parse.quote(item_id, safe="")


# IS-10192: from https://carsonip.me/posts/azure-tcp-idle-timeout-tcp-keepalive-python/
class HTTPAdapterWithSocketOptions(requests.adapters.HTTPAdapter):
    def __init__(self, *args, **kwargs):
        if "socket_options" in kwargs:
            self.socket_options = kwargs.pop("socket_options", None)
        super(HTTPAdapterWithSocketOptions, self).__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        if self.socket_options is not None:
            kwargs["socket_options"] = self.socket_options

        super(HTTPAdapterWithSocketOptions, self).init_poolmanager(*args, **kwargs)


class HTTPAdapterWithTCPKeepalive(HTTPAdapterWithSocketOptions):
    def __init__(self, *args, **kwargs):
        self.socket_options = HTTPConnection.default_socket_options + [
                                (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
                            ]

        # Set the default keepalive params explicitly, the following values are equivalent to:
        # net.ipv4.tcp_keepalive_time = 180
        # net.ipv4.tcp_keepalive_intvl = 15
        # net.ipv4.tcp_keepalive_probes = 5

        TCP_KEEPIDLE = 180
        TCP_KEEPINTVL = 15
        TCP_KEEPCNT = 5

        if getattr(socket, 'TCP_KEEPINTVL', None) is not None:
            self.socket_options += [(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, TCP_KEEPINTVL)]
        elif sys.platform == 'darwin':
            TCP_KEEPALIVE = getattr(socket, 'TCP_KEEPALIVE', 0x10)
            self.socket_options += [(socket.IPPROTO_TCP, TCP_KEEPALIVE, TCP_KEEPINTVL)]

        if getattr(socket, 'TCP_KEEPCNT', None) is not None:
            self.socket_options += [(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, TCP_KEEPCNT)]

        if getattr(socket, 'TCP_KEEPIDLE', None) is not None:
            self.socket_options += [(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, TCP_KEEPIDLE)]

        # See https://blog.cloudflare.com/when-tcp-sockets-refuse-to-die/
        #if getattr(socket, 'TCP_USER_TIMEOUT', None) is not None:
        #    self.socket_options += [(socket.IPPROTO_TCP, socket.TCP_USER_TIMEOUT, TCP_KEEPIDLE + TCP_KEEPINTVL * TCP_KEEPCNT)]

        super(HTTPAdapterWithTCPKeepalive, self).__init__(*args, **kwargs)


class Connection:
    """This class represents a connection to a Sesam installation. This is the starting point of all interactions
    with the Sesam installation.
    """
    def __init__(self, sesamapi_base_url,
                 username=None, password=None,
                 client_certificate=None,
                 timeout=30,
                 retries=0,
                 retry_delay=5,
                 jwt_auth_token=None,
                 client_id=None, client_secret=None, token_endpoint=None,
                 verify_ssl=True):
        """
        client_id, client_secret and token_endpoint are specified when we want to use the OpenID connect
        "client credentials" grant-type.
        """
        self._close_has_been_called_event = threading.Event()

        if not sesamapi_base_url.endswith("/"):
            sesamapi_base_url += "/"

        self.sesamapi_base_url = sesamapi_base_url
        self.username = username
        self.password = password
        self._client_certificate = client_certificate
        self.timeout = timeout
        self._jwt_auth_token = jwt_auth_token
        self._verify_ssl = verify_ssl
        self._retries = retries
        self._retry_delay = retry_delay

        self._headers = {
            "ACCEPT": "application/json,*/*"
        }

        self._oauth2_config = None
        if client_id:
            self._oauth2_config = {
                "client_id": client_id,
                "client_secret": client_secret,
                "token_endpoint": token_endpoint,
            }

        self._oauth2_client = None
        self._oauth2_session = None
        self._oauth2_token = None
        self._session = None
        self._adapter = HTTPAdapterWithTCPKeepalive()

    def set_jwt_auth_token(self, jwt_auth_token):
        """Replaces the current jwt_auth_token (if any) with the new jwt."""
        self._jwt_auth_token = jwt_auth_token
        # We must close any existing session, since the old session uses the old jwt token. A new session will
        # be lazily created by the session() method when needed.
        session = self._session
        self._session = None
        if session is not None:
            session.close()

    @property
    def session(self):
        if self._oauth2_config is not None:
            oauth2_config = self._oauth2_config

            if self._oauth2_client is None:
                self._oauth2_client = BackendApplicationClient(client_id=oauth2_config["client_id"])

            if self._oauth2_session is None:
                self._oauth2_session = OAuth2Session(client=self._oauth2_client)
                if self._headers:
                    self._oauth2_session.headers = self._headers

                # IS-10192: turn on TCP keep-alive to avoid indefinitely hanging connections
                self._oauth2_session.mount("http://", self._adapter)
                self._oauth2_session.mount("https://", self._adapter)

            # If no token has been created yet or if the previous token has expired (or is close to it),
            # fetch a new access token before returning the session to the callee.
            if self._oauth2_token is None or self._oauth2_token["expires_at"] <= (time.time() + 60):
                self._oauth2_token = self._oauth2_session.fetch_token(
                    token_url=oauth2_config["token_endpoint"],
                    client_id=oauth2_config["client_id"],
                    client_secret=oauth2_config["client_secret"],
                )

            return self._oauth2_session
        else:
            if self._session is None:
                if self._jwt_auth_token:
                    self._headers["AUTHORIZATION"] = "Bearer %s" % (self._jwt_auth_token,)

                self._session = session = requests.Session()
                session.verify = self._verify_ssl

                if self.username:
                    auth = (self.username, self.password)
                else:
                    auth = None

                if auth is not None:
                    session.auth = auth
                if self._client_certificate is not None:
                    session.cert = self._client_certificate
                session.headers = self._headers

                # IS-10192: turn on TCP keep-alive to avoid indefinitely hanging connections
                self._session.mount("http://", self._adapter)
                self._session.mount("https://", self._adapter)
            return self._session

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        # IS-8849: interrupt any pending do_* retries
        self._close_has_been_called_event.set()

        if self._session is not None:
            self._session.close()
        self._session = None

        if self._oauth2_session is not None:
            self._oauth2_session.close()
        self._oauth2_session = None
        self._oauth2_token = None
        self._oauth2_client = None

    def _get_requests_kwargs(self):
        kwargs = {}
        if self.timeout:
            kwargs["timeout"] = self.timeout
        return kwargs

    def do_put_request(self, url, data=None, allowable_response_status_codes=(200,), retries=None, retry_delay=None,
                       files=None,
                       **kwargs):
        assert url is not None
        session_put_kwargs = self._get_requests_kwargs()
        session_put_kwargs.update(kwargs)

        if retries is None:
            retries = self._retries

        if retry_delay is None:
            retry_delay = self._retry_delay

        while True:
            try:
                try:
                    response = self.session.put(url, data=data, files=files, **session_put_kwargs)
                except TokenExpiredError:
                    # Manually clear the token and retry once. The token is supposed to be automatically
                    # refreshed, but if the local clock is off that can mess things up.
                    self._oauth2_token = None
                    self._rewind_request_data_if_needed(data, files)
                    response = self.session.put(url, data=data, **session_put_kwargs)

                if allowable_response_status_codes:
                    utils.validate_response_is_ok(response, allowable_response_status_codes)

                return response
            except (RequestException, ConnectionError,
                    InternalServerError, ServiceUnavailable, BadRequestException) as e:
                if retries <= 0:
                    raise e

                self._rewind_request_data_if_needed(data, files)

                retries -= 1
                self.interruptable_sleep(retry_delay)

    def do_get_request(self, url, allowable_response_status_codes=(200, 404), retries=None, retry_delay=None, **kwargs):
        assert url is not None
        session_get_kwargs = self._get_requests_kwargs()
        session_get_kwargs.update(kwargs)

        if retries is None:
            retries = self._retries

        if retry_delay is None:
            retry_delay = self._retry_delay

        while True:
            try:
                try:
                    response = self.session.get(url, **session_get_kwargs)
                except TokenExpiredError:
                    # Manually clear the token and retry once. The token is supposed to be automatically
                    # refreshed, but if the local clock is off that can mess things up.
                    self._oauth2_token = None
                    response = self.session.get(url, **session_get_kwargs)

                if allowable_response_status_codes:
                    utils.validate_response_is_ok(response, allowable_response_status_codes)

                return response
            except (RequestException, ConnectionError,
                    InternalServerError, ServiceUnavailable, BadRequestException) as e:
                if retries <= 0:
                    raise e

                retries -= 1
                self.interruptable_sleep(retry_delay)

    def do_head_request(self, url, allowable_response_status_codes=(200, 404), retries=None, retry_delay=None, **kwargs):
        assert url is not None
        session_get_kwargs = self._get_requests_kwargs()
        session_get_kwargs.update(kwargs)

        if retries is None:
            retries = self._retries

        if retry_delay is None:
            retry_delay = self._retry_delay

        while True:
            try:
                try:
                    response = self.session.head(url, **session_get_kwargs)
                except TokenExpiredError:
                    # Manually clear the token and retry once. The token is supposed to be automatically
                    # refreshed, but if the local clock is off that can mess things up.
                    self._oauth2_token = None
                    response = self.session.get(url, **session_get_kwargs)

                if allowable_response_status_codes:
                    utils.validate_response_is_ok(response, allowable_response_status_codes)

                return response
            except (RequestException, ConnectionError,
                    InternalServerError, ServiceUnavailable, BadRequestException) as e:
                if retries <= 0:
                    raise e

                retries -= 1
                self.interruptable_sleep(retry_delay)

    def do_options_request(self, url, allowable_response_status_codes=(200, 404), retries=None, retry_delay=None,
                           **kwargs):
        assert url is not None
        session_get_kwargs = self._get_requests_kwargs()
        session_get_kwargs.update(kwargs)

        if retries is None:
            retries = self._retries

        if retry_delay is None:
            retry_delay = self._retry_delay

        while True:
            try:
                try:
                    response = self.session.options(url, **session_get_kwargs)
                except TokenExpiredError:
                    # Manually clear the token and retry once. The token is supposed to be automatically
                    # refreshed, but if the local clock is off that can mess things up.
                    self._oauth2_token = None
                    response = self.session.get(url, **session_get_kwargs)

                if allowable_response_status_codes:
                    utils.validate_response_is_ok(response, allowable_response_status_codes)

                return response
            except (RequestException, ConnectionError,
                    InternalServerError, ServiceUnavailable, BadRequestException) as e:
                if retries <= 0:
                    raise e

                retries -= 1
                self.interruptable_sleep(retry_delay)

    def do_post_request(self, url, allowable_response_status_codes=(200,), retries=None, retry_delay=None,
                        data=None,
                        files=None,
                        **kwargs):
        assert url is not None
        session_post_kwargs = self._get_requests_kwargs()
        session_post_kwargs.update(kwargs)

        if retries is None:
            retries = self._retries

        if retry_delay is None:
            retry_delay = self._retry_delay

        while True:
            try:
                try:
                    response = self.session.post(url, data=data, files=files, **session_post_kwargs)
                except TokenExpiredError:
                    # Manually clear the token and retry once. The token is supposed to be automatically
                    # refreshed, but if the local clock is off that can mess things up.
                    self._oauth2_token = None
                    self._rewind_request_data_if_needed(data, files)
                    response = self.session.post(url, data=data, **session_post_kwargs)

                if allowable_response_status_codes:
                    utils.validate_response_is_ok(response, allowable_response_status_codes)

                return response
            except (RequestException, ConnectionError,
                    InternalServerError, ServiceUnavailable, BadRequestException) as e:
                if retries <= 0:
                    raise e

                self._rewind_request_data_if_needed(data, files)

                retries -= 1
                self.interruptable_sleep(retry_delay)

    def _rewind_request_data_if_needed(self, data, files):
        if isinstance(data, io.BytesIO):
            data.seek(0)

        if files:
            for name, fileinfo in files.items():
                if type(fileinfo) is tuple:
                    # This is a filetuple on the form ('filename', fileobj)
                    fileobj = fileinfo[1]
                else:
                    fileobj = fileinfo
                if isinstance(fileobj, io.BytesIO):
                    fileobj.seek(0)

    def do_delete_request(self, url, allowable_response_status_codes=(200,), retries=None, retry_delay=None, **kwargs):
        assert url is not None
        session_delete_kwargs = self._get_requests_kwargs()
        session_delete_kwargs.update(kwargs)

        if retries is None:
            retries = self._retries

        if retry_delay is None:
            retry_delay = self._retry_delay

        while True:
            try:
                try:
                    response = self.session.delete(url, **session_delete_kwargs)
                except TokenExpiredError:
                    # Manually clear the token and retry once. The token is supposed to be automatically
                    # refreshed, but if the local clock is off that can mess things up.
                    self._oauth2_token = None
                    response = self.session.delete(url, **session_delete_kwargs)

                if allowable_response_status_codes:
                    utils.validate_response_is_ok(response, allowable_response_status_codes)

                return response
            except (RequestException, ConnectionError,
                    InternalServerError, ServiceUnavailable, BadRequestException) as e:
                if retries <= 0:
                    raise e

                retries -= 1
                self.interruptable_sleep(retry_delay)

    def interruptable_sleep(self, sleep_time):
        self._close_has_been_called_event.wait(sleep_time)

    @property
    def pipes_url(self):
        return self.sesamapi_base_url + "pipes"

    def get_pipe_url(self, pipe_id):
        return self.pipes_url + "/" + quote_id_for_url(pipe_id)

    def get_pipe_config_url(self, pipe_id):
        return self.get_pipe_url(pipe_id) + "/config"

    def get_pipe_schema_def_url(self, pipe_id, sample_size, keys_only):
        return self.get_pipe_url(pipe_id) + "/generate-schema-definition?sample_size=%s&keys_only=%s" % (sample_size,
                                                                                                         keys_only)

    @property
    def logs_url(self):
        return self.sesamapi_base_url + "logs"

    def get_log_url(self, log_id):
        return self.logs_url + "/" + quote_id_for_url(log_id)

    def get_pipe_pump_url(self, pipe_id):
        return self.get_pipe_url(pipe_id) + "/pump"

    def get_pipe_entities_url(self, pipe_id):
        return self.get_pipe_url(pipe_id) + "/entities"

    def get_pipe_log_url(self, pipe_id):
        return self.get_pipe_url(pipe_id) + "/log"

    def get_pipe_metadata_url(self, pipe_id):
        return self.get_pipe_url(pipe_id) + "/metadata"

    def get_pipe_receiver_endpoint_url(self, pipe_id):
        return self.sesamapi_base_url + "receivers/" + quote_id_for_url(pipe_id) + "/entities"

    @property
    def systems_url(self):
        return self.sesamapi_base_url + "systems"

    def get_system_url(self, system_id):
        return self.systems_url + "/" + quote_id_for_url(system_id)

    def get_system_config_url(self, system_id):
        return self.get_system_url(system_id) + "/config"

    def get_system_status_url(self, system_id):
        return self.get_system_url(system_id) + "/status"

    def get_system_metadata_url(self, system_id):
        return self.get_system_url(system_id) + "/metadata"

    def get_system_source_prototypes_url(self, system_id):
        return self.get_system_url(system_id) + "/source-prototypes"

    def get_system_sink_prototypes_url(self, system_id):
        return self.get_system_url(system_id) + "/sink-prototypes"

    @property
    def systemconfigs_url(self):
        return self.sesamapi_base_url + "system-prototypes"

    @property
    def models_url(self):
        return self.sesamapi_base_url + "models"

    def get_model_url(self, model_id):
        return self.models_url + "/" + quote_id_for_url(model_id)

    @property
    def stats_url(self):
        return self.sesamapi_base_url + "stats"

    @property
    def secrets_url(self):
        return self.sesamapi_base_url + "secrets"

    @property
    def secrets_values_url(self):
        return self.sesamapi_base_url + "secrets_values"

    def get_secrets_url(self, system_id):
        return self.get_system_url(system_id) + "/secrets"

    def get_secrets_values_url(self, system_id):
        return self.get_system_url(system_id) + "/secrets_values"

    @property
    def env_vars_url(self):
        return self.sesamapi_base_url + "env"

    @property
    def status_url(self):
        return self.sesamapi_base_url + "status"

    @property
    def ping_url(self):
        return self.sesamapi_base_url + "ping"

    @property
    def admin_url(self):
        return self.sesamapi_base_url + "admin"

    @property
    def node_metadata_url(self):
        return self.sesamapi_base_url + "metadata"

    @property
    def node_namespaces_url(self):
        return self.sesamapi_base_url + "namespaces"

    def get_secret_url(self, key, system_id=None):
        if system_id is None:
            return self.secrets_url + "/" + quote_id_for_url(key)
        else:
            return self.get_secrets_url(system_id) + "/" + quote_id_for_url(key)

    def get_env_var_url(self, key):
        return self.env_vars_url + "/" + quote_id_for_url(key)

    @property
    def node_license_url(self):
        return self.sesamapi_base_url + "license"

    @property
    def datasets_url(self):
        return self.sesamapi_base_url + "datasets"

    def get_config_url(self, config_group=None):
        url = self.sesamapi_base_url + "config"
        if config_group is not None:
            url += "/" + urllib.parse.quote(config_group, safe="")
        return url

    def get_dataset_url(self, dataset_id):
        return self.datasets_url + "/" + quote_id_for_url(dataset_id)

    def get_dataset_entities_url(self, dataset_id):
        return self.get_dataset_url(dataset_id) + "/entities"

    def get_dataset_entity_url(self, dataset_id):
        return self.get_dataset_url(dataset_id) + "/entity"

    def get_dataset_metadata_url(self, dataset_id):
        return self.get_dataset_url(dataset_id) + "/metadata"

    def get_dataset_indexes_url(self, dataset_id):
        return self.get_dataset_url(dataset_id) + "/indexes"

    def restart_node(self, **kwargs):
        """
            restarts the node via the admin api, it returns a dict {"message": "OK"} if successful
        """

        url = self.admin_url

        postdata = {"operation": "restart"}

        _allowable_response_status_codes = [200]
        restart_response = self.do_post_request(url, allowable_response_status_codes=_allowable_response_status_codes,
                                                data=postdata, **kwargs)

        validate_response_is_ok(restart_response, allowable_response_status_codes=_allowable_response_status_codes)
        return utils.parse_json_response(restart_response)

    def reset_node(self, **kwargs):
        """
            resets the node via the admin api, it returns a dict {"message": "OK"} if successful
        """

        url = self.admin_url

        postdata = {"operation": "reset"}

        _allowable_response_status_codes = [200]
        restart_response = self.do_post_request(url, allowable_response_status_codes=_allowable_response_status_codes,
                                                data=postdata, **kwargs)

        validate_response_is_ok(restart_response, allowable_response_status_codes=_allowable_response_status_codes)
        return utils.parse_json_response(restart_response)

    def get_systems(self, include_internal_systems=False, **kwargs):
        systems = []

        # TODO: wrap in a utility-method and add error-detection
        request_params = {}
        if include_internal_systems:
            request_params["include-internal-systems"] = "true"
        response = self.do_get_request(self.systems_url, allowable_response_status_codes=[200],
                                       params=request_params, **kwargs)
        parsed_response = utils.parse_json_response(response)

        for system_json in parsed_response:
            systems.append(System(self, system_json))

        return systems

    def get_system(self, system_id, **kwargs):
        response = self.do_get_request(self.get_system_url(system_id), **kwargs)
        if response.status_code == 404:
            return None
        system_json = utils.parse_json_response(response)
        return System(self, system_json)

    def get_models(self, **kwargs):
        models = []

        response = self.do_get_request(self.models_url, allowable_response_status_codes=[200], **kwargs)
        parsed_response = utils.parse_json_response(response)

        for model_json in parsed_response:
            models.append(Model(self, model_json))

        return models

    def get_model(self, model_id, **kwargs):
        response = self.do_get_request(self.get_model_url(model_id), **kwargs)
        if response.status_code == 404:
            return None
        model_json = utils.parse_json_response(response)
        return Model(self, model_json)

    def get_systemconfigs(self, **kwargs):
        response = self.do_get_request(self.systemconfigs_url, allowable_response_status_codes=[200], **kwargs)
        return utils.parse_json_response(response)

    def get_stats(self, **kwargs):
        """
        :return: A nested dict with metrics.
        """
        stats_response = self.do_get_request(self.stats_url, allowable_response_status_codes=[200], **kwargs)
        return utils.parse_json_response(stats_response)

    def get_status(self, **kwargs):
        """
        :return: A dict with various status-information (disk-usage, etc.)
        """
        status_response = self.do_get_request(self.status_url, **kwargs)
        return utils.parse_json_response(status_response)

    def get_ping(self, **kwargs):
        """
        :return: A dict with {"status": "running"} if successful
        """
        status_response = self.do_get_request(self.ping_url, **kwargs)
        return utils.parse_json_response(status_response)

    def get_datasets(self, include_internal_datasets=True, verbose=False, **kwargs):
        # Note: the default for getting datasets is to get all
        # including the internal ones
        # This is opposite behaviour to pipes/systems
        datasets = []
        params = {}
        if not include_internal_datasets:
            params["include-internal-datasets"] = include_internal_datasets
        if verbose:
            params["verbose"] = verbose
        response = self.do_get_request(self.datasets_url, allowable_response_status_codes=[200],
                                       params=params, **kwargs)
        parsed_response = utils.parse_json_response(response)
        for dataset_json in parsed_response:
            datasets.append(Dataset(self, dataset_json))
        return datasets

    def get_dataset(self, dataset_id, **kwargs):
        response = self.do_get_request(self.get_dataset_url(dataset_id), **kwargs)
        if response.status_code == 404:
            return None
        dataset_json = utils.parse_json_response(response)
        return Dataset(self, dataset_json)

    def get_pipes(self, include_internal_pipes=False, fields=None, **kwargs):
        pipes = []

        # TODO: wrap in a utility-method and add error-detection
        request_params = {}
        if include_internal_pipes:
            request_params["include-internal-pipes"] = "true"

        if fields is not None:
            request_params["fields"] = fields
        response = self.do_get_request(self.pipes_url,
                                       allowable_response_status_codes=[200],
                                       params=request_params, **kwargs
                                       )

        parsed_response = utils.parse_json_response(response)

        for pipe_json in parsed_response:
            pipes.append(Pipe(self, pipe_json))
        return pipes

    def get_pipe(self, pipe_id, fields=None, **kwargs):
        # TODO: wrap in a utility-method and add error-detection

        request_params = kwargs.pop("request_params", {})

        if fields is not None:
            request_params["fields"] = ",".join(fields)
        response = self.do_get_request(self.get_pipe_url(pipe_id), params=request_params, **kwargs)
        if response.status_code == 404:
            return None
        pipe_json = utils.parse_json_response(response)
        return Pipe(self, pipe_json)

    def get_pipe_schema_def(self, pipe_id, sample_size=1000, keys_only=False, **kwargs):
        # TODO: wrap in a utility-method and add error-detection
        response = self.do_get_request(self.get_pipe_schema_def_url(pipe_id, sample_size, keys_only), **kwargs)
        if response.status_code == 404:
            return None

        return utils.parse_json_response(response)

    def add_pipes(self, pipe_configs, force=False, **kwargs):
        request_params = {"force": "true" if force else "false"}
        response = self.do_post_request(self.pipes_url, allowable_response_status_codes=[201, 409, 400],
                                        json=pipe_configs, params=request_params, **kwargs)

        if response.status_code == 409:
            raise PipeAlreadyExists(response.text)

        parsed_response = utils.parse_json_response(response)

        if response.status_code == 400:
            raise ConfigUploadFailed(response=response)

        pipes = [Pipe(self, pipe_json) for pipe_json in parsed_response]
        return pipes

    def add_systems(self, system_configs, force=False, **kwargs):
        request_params = {"force": "true" if force else "false"}
        response = self.do_post_request(self.systems_url, allowable_response_status_codes=[201, 409, 400],
                                        json=system_configs, params=request_params, **kwargs)

        if response.status_code == 409:
            raise SystemAlreadyExists(response.text)

        parsed_response = utils.parse_json_response(response)

        if response.status_code == 400:
            raise ConfigUploadFailed(response=response)

        systems = [System(self, system_json) for system_json in parsed_response]
        return systems

    def add_models(self, model_configs, **kwargs):
        response = self.do_post_request(self.models_url, allowable_response_status_codes=[201, 409, 400],
                                        json=model_configs, **kwargs)

        parsed_response = utils.parse_json_response(response)

        if response.status_code == 400:
            raise ConfigUploadFailed(response=response)

        models = [Model(self, model_json) for model_json in parsed_response]
        return models

    def get_logs(self, **kwargs):
        logs = []

        # TODO: wrap in a utility-method and add error-detection
        response = self.do_get_request(self.logs_url, allowable_response_status_codes=[200], **kwargs)
        parsed_response = utils.parse_json_response(response)

        for log_json in parsed_response:
            logs.append(Log(self, log_json))
        return logs

    def get_log_content(self, log_id, **kwargs):
        """This returns a stream with the content of the specified logfile"""
        # TODO: wrap in a utility-method and add error-detection
        response = self.do_get_request(self.get_log_url(log_id), stream=True, **kwargs)
        if response.status_code == 404:
            return None
        return response.raw

    def get_metadata(self, **kwargs):
        """Gets the current metadata for the node"""
        response = self.do_get_request(self.node_metadata_url, **kwargs)
        return utils.parse_json_response(response)

    def set_metadata(self, metadata, **kwargs):
        """Replaces the metadata for the node with the specified dictionary"""
        self.do_put_request(self.node_metadata_url, allowable_response_status_codes=[200], json=metadata, **kwargs)

    def get_namespaces(self, **kwargs):
        """Gets the current metadata for the node"""
        response = self.do_get_request(self.node_namespaces_url, allowable_response_status_codes=[200], **kwargs)
        return utils.parse_json_response(response)

    def get_secrets(self, include_internal_secrets=False, **kwargs):
        """This returns a list with the names of the secrets."""
        request_params = {}
        if include_internal_secrets:
            request_params["include-internal-secrets"] = "true"

        response = self.do_get_request(self.secrets_url, allowable_response_status_codes=[200], params=request_params,
                                       **kwargs)
        secrets_info = utils.parse_json_response(response)
        return secrets_info

    def get_secrets_values(self, **kwargs):
        """This returns a dict with secretname=>secretvalue mappings. This is only available for nodes that
        are running in ci-test mode. In addition, the user has to have the "SystemAdmin" role."""
        request_params = {}
        response = self.do_get_request(self.secrets_values_url, allowable_response_status_codes=[200],
                                       params=request_params, **kwargs)
        secrets_info = utils.parse_json_response(response)
        return secrets_info

    def put_secrets(self, secrets_content, dont_encrypt=False, **kwargs):
        encrypt = dont_encrypt and "false" or "true"
        response = self.do_put_request(self.secrets_url, allowable_response_status_codes=[200],
                                       json=secrets_content, params=dict(encrypt=encrypt), **kwargs)

        secrets_info = utils.parse_json_response(response)
        return secrets_info

    def post_secrets(self, secrets_content, dont_encrypt=False, **kwargs):
        encrypt = dont_encrypt and "false" or "true"
        response = self.do_post_request(self.secrets_url, allowable_response_status_codes=[200],
                                        json=secrets_content, params=dict(encrypt=encrypt), **kwargs)

        secrets_info = utils.parse_json_response(response)
        return secrets_info

    def delete_secret(self, key, **kwargs):
        response = self.do_delete_request(self.get_secret_url(key), allowable_response_status_codes=[200], **kwargs)
        secrets_info = utils.parse_json_response(response)
        return secrets_info

    def get_env_vars(self, **kwargs):
        response = self.do_get_request(self.env_vars_url, allowable_response_status_codes=[200], **kwargs)
        env_vars_info = utils.parse_json_response(response)
        return env_vars_info

    def put_env_vars(self, env_vars_content, **kwargs):
        response = self.do_put_request(self.env_vars_url, allowable_response_status_codes=[200],
                                       json=env_vars_content, **kwargs)

        env_vars_info = utils.parse_json_response(response)
        return env_vars_info

    def post_env_vars(self, env_vars_content, **kwargs):
        response = self.do_post_request(self.env_vars_url, allowable_response_status_codes=[200], json=env_vars_content,
                                        **kwargs)

        env_vars_info = utils.parse_json_response(response)
        return env_vars_info

    def delete_env_var(self, key, **kwargs):
        response = self.do_delete_request(self.get_env_var_url(key), allowable_response_status_codes=[200], **kwargs)
        env_vars_info = utils.parse_json_response(response)
        return env_vars_info

    def get_license(self, **kwargs):
        """Gets the current license info for the node"""
        response = self.do_get_request(self.node_license_url, **kwargs)
        license_info = utils.parse_json_response(response)
        return license_info

    def set_license(self, license_key_content, **kwargs):
        """Replaces the license for the node with the specified license key"""
        response = self.do_put_request(self.node_license_url, allowable_response_status_codes=[200],
                                       data=license_key_content, **kwargs)
        license_info = utils.parse_json_response(response)
        return license_info

    def delete_metadata(self, **kwargs):
        """Deleted all metadata for the node"""
        self.do_delete_request(self.node_metadata_url, allowable_response_status_codes=[200], **kwargs)

    def get_config(self, config_group=None, **kwargs):
        """Returns the configuration of the sesam node.
        :param config_group: The config-group to get the config from. The default is None, meaning the main config
                      of the node.
        """
        response = self.do_get_request(self.get_config_url(config_group), allowable_response_status_codes=[200],
                                       **kwargs)
        utils.validate_equal_case_insensitive(response.headers["Content-Type"], "application/json")
        config = utils.parse_json_response(response)
        return config

    def get_config_history(self, entity_id, since=None, limit=None, reverse=None, **kwargs):
        """Returns the history of the configuration entity of the sesam node.

        :param entity_id: the id of the entity

        :param since: The opaque value that can be used to efficiently skip already seen entities.

        :param limit: The "limit" parameter specifies the maximum number of entities to return.
                      If this is not specified, all available entities are returned.

        :param reverse: If this is False (the default) entities will returned with the oldest entity first. \
                        If this is True, the newest entity will be returned first.
        """

        url = self.sesamapi_base_url + "config-history/entity"

        request_params = {"entity_id": entity_id}

        if since is not None:
            request_params["since"] = since

        if limit is not None:
            request_params["limit"] = limit

        if reverse and reverse is True:
            request_params["reverse"] = reverse

        response = self.do_get_request(url, allowable_response_status_codes=[200], params=request_params, **kwargs)
        utils.validate_equal_case_insensitive(response.headers["Content-Type"], "application/json")
        config = utils.parse_json_response(response)

        return config

    def get_config_as_json_string(self, config_group=None, **kwargs):
        """Returns the configuration of the sesam node as a json-string, using the same formatting
        as the raw server response.
        :param config_group: The config-group to get the config from. The default is None, meaning the main config
                      of the node.
        """
        response = self.do_get_request(self.get_config_url(config_group), allowable_response_status_codes=[200],
                                       **kwargs)
        utils.validate_equal_case_insensitive(response.headers["Content-Type"], "application/json")
        return response.content.decode("utf-8")

    def get_config_as_zip(self, config_group=None, **kwargs):
        """Returns the configuration of the sesam node as a zip-archive with separate files for each pipe, system,
        etc.
        :param config_group: The config-group to get the config from. The default is None, meaning the main config
                      of the node.

        Tip: the returned bytes can be parsed like this:
               config = zipfile.ZipFile(io.BytesIO(returned_bytes)
        """
        response = self.do_get_request(self.get_config_url(config_group),
                                       allowable_response_status_codes=[200],
                                       headers={"ACCEPT": "application/zip"}, **kwargs
                                       )
        utils.validate_equal_case_insensitive(response.headers["Content-Type"], "application/zip")
        return response.content

    def upload_config(self, config,
                      config_group=None,
                      force=False,
                      _use_multipart_form_data=True, **kwargs):
        """
        Uploads the specified configuration. The new configuration will replace any existing
        configuration. Any existing configuration entities that are not in the new configuration will
        be removed.

        :param config: The config. This can be either a filename, a file-like object, or a bytes-object. The format
                      must be either a json object (as returned by the get_config()-method, or a zip-archive (as
                      returned by the get_config_as_zip()-method).
        :param config_group: The config-group to upload the config to. The default is None, meaning the main config
                      of the node.
        :param force: If this is False (the default), no changes will be made if the configuration contains any errors.
                      If it is True, the node will attempt to apply the new configuration on a best-effort basis, and
                      component with erronous configuration will get their "is-valid-config" flag set to false.
        :param _use_multipart_form_data: This can be set to False in order to post the config content as the
                                request-body, instead of posting it using the normal "multipart/form-data" encoding.
                                This is only used in ci-tests.

        :returns The parsed response, which will contain a "validation_errors" list. This list will be empty if the
                 upload completed with no errors or warnings.
        """
        theconfigfile_must_be_closed = False

        if isinstance(config, (dict, list)):
            # The configuration is in the form of a json object
            theconfigfile = io.BytesIO(json.dumps(config).encode("utf-8"))

        elif isinstance(config, io.IOBase):
            # The input is a file-like object
            theconfigfile = config
            theconfigfile.seek(0)

        elif isinstance(config, (bytes, bytearray)):
            # The input is the bytes of the zipfile
            theconfigfile = io.BytesIO(config)

        elif isinstance(config, str):
            # The input is a filename
            theconfigfile = open(config, "rb")
            theconfigfile_must_be_closed = True

        else:
            raise TypeError(
                "The 'config' parameter must be json-object, json-array, a file-like object or a filename, but it was of type '%s'" % (
                    type(config),))

        # Try to figure out the content-type to use.
        try:
            zipfile.ZipFile(theconfigfile)
            content_type = 'application/zip'
        except zipfile.BadZipfile:
            try:
                theconfigfile.seek(0)
                json.loads(theconfigfile.read().decode("utf-8"))
                content_type = 'application/json'
            except Exception as e:
                raise TypeError(
                    "The config must be either a zip-archive or a json object! The specified config was neither: %s"
                        % (e,))

        theconfigfile.seek(0)

        request_params = {"force": "true" if force else "false"}

        try:
            if _use_multipart_form_data:
                # This is the normal case. We use the same method a browser would use when uploading a file.
                files = {'file': ('config', theconfigfile, content_type)}
                response = self.do_put_request(self.get_config_url(config_group),
                                               params=request_params,
                                               # we will deal with any 400-errors later
                                               allowable_response_status_codes=[200, 400],
                                               files=files, **kwargs)
            else:
                # We want to put the config content directly in the response body.
                data = theconfigfile.read()
                if content_type == 'application/zip':
                    if not isinstance(data, bytes):
                        raise TypeError(
                            "The content-type is 'application/zip', but the data isn't of type 'bytes': type(data):%s"
                                % (type(data),))
                response = self.do_put_request(self.get_config_url(config_group), data=data,
                                               params=request_params,
                                               # we will deal with any 400-errors later
                                               allowable_response_status_codes=[200, 400],
                                               headers={"Content-Type": content_type}, **kwargs)

        finally:
            if theconfigfile_must_be_closed:
                theconfigfile.close()

        utils.validate_equal_case_insensitive(response.headers["Content-Type"], "application/json")

        parsed_response = utils.parse_json_response(response)

        if response.status_code == 400:
            raise ConfigUploadFailed(response=response)

        return parsed_response

    def get_access_control_list(self, acl_path, do_transit_decoding=True, **kwargs):
        """Retrieves the specified acl. The acl_path is the relative part of the url to the acl.

        Examples:
            "dataset_collection"
            "dataset/some_dataset_id"
            "dataset_prototype"
        """
        acl_url = self.sesamapi_base_url + "permissions/" + acl_path
        response = self.do_get_request(acl_url, allowable_response_status_codes=[200], **kwargs)
        utils.validate_equal_case_insensitive(response.headers["Content-Type"], "application/json")
        acl = entity_json.parse_entity(response.content.decode("utf-8"), do_transit_decoding=do_transit_decoding)
        return acl

    def set_access_control_list(self, acl_path, acl, **kwargs):
        """Uploads the specified acl. The acl_path is the relative part of the url to the acl (see
        the get_access_control_list() method for examples.
        """
        acl_url = self.sesamapi_base_url + "permissions/" + acl_path
        self.do_put_request(acl_url, allowable_response_status_codes=[200], json=acl, **kwargs)

    def login(self, email, password, **kwargs):
        login_url = self.sesamapi_base_url + "jwt"

        response = self.do_post_request(login_url, allowable_response_status_codes=[200],
                                        data={"email": email, "password": password}, **kwargs)

        jwt = response.text

        return jwt

    def get_backup(self, backup_id, **kwargs):
        url = self.sesamapi_base_url + "backups/" + backup_id
        response = self.do_get_request(url, allowable_response_status_codes=[200], **kwargs)
        backup_info = response.json()
        return backup_info

    def get_backups(self, **kwargs):
        url = self.sesamapi_base_url + "backups"
        response = self.do_get_request(url, allowable_response_status_codes=[200], **kwargs)
        backups = response.json()
        return backups

    def trigger_backup(self, **kwargs):
        url = self.sesamapi_base_url + "backups"
        response = self.do_post_request(url, allowable_response_status_codes=[200], **kwargs)
        backup_info = response.json()
        return backup_info

    def wait_for_backup_to_finish(self, backup_id, timeout=60, poll_interval=0.5, **kwargs):
        """Utility method for waiting until the specified backup is no longer running.

        :param timeout: The maximum time to wait for the backup to stop running. If the timeout is exceeded,
                        an error is raised.
        """
        starttime = time.monotonic()
        while True:
            url = self.sesamapi_base_url + "backups/" + backup_id
            response = self.do_get_request(url, allowable_response_status_codes=[200], **kwargs)
            backup_info = response.json()
            if backup_info["status"] != "running":
                break
            elapsedtime = time.monotonic() - starttime
            if elapsedtime > timeout:
                raise TimeoutWhileWaitingForBackupToFinishException(
                    "Timed out while waiting for the backup to finish running! lat backup_info:\n%s" % (
                        backup_info,))
            time.sleep(poll_interval)

    def trigger_restore(self, node_id_regexp, target_backup_time):
        url = self.sesamapi_base_url + "restores"
        response = self.do_post_request(url, allowable_response_status_codes=[200],
                                        data={
                                            "node_id_regexp": node_id_regexp,
                                            "target_backup_time": target_backup_time,
                                        })
        restore_info = response.json()
        return restore_info

    def wait_for_restore_to_finish(self, restore_id, timeout=600, poll_interval=0.5, **kwargs):
        """Utility method for waiting until the specified restore is no longer running.

        :param timeout: The maximum time to wait for the restore to stop running. If the timeout is exceeded,
                        an error is raised.
        """
        url = self.sesamapi_base_url + "restores/" + restore_id
        starttime = time.monotonic()
        while True:
            try:
                response = self.do_get_request(url, allowable_response_status_codes=[200], **kwargs)
            except Exception:
                # It is normal that the api stops responding during a restore, so we can ignore this
                # exception unless we time out.
                elapsedtime = time.monotonic() - starttime
                if elapsedtime > timeout:
                    raise
                time.sleep(poll_interval)
            restore_info = response.json()
            if restore_info["status"] == "success":
                return restore_info
            elapsedtime = time.monotonic() - starttime
            if elapsedtime > timeout:
                raise TimeoutWhileWaitingForRestoreToFinishException(
                    "Timed out while waiting for the restore to finish running! restore_info:\n%s" % (
                        restore_info,))

            time.sleep(poll_interval)

    def get_api_info(self, **kwargs):
        response = self.do_get_request(self.sesamapi_base_url, allowable_response_status_codes=[200], **kwargs)
        info = response.json()
        return info



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import os
import json

from ..utils import print_item, exit_with_error


def get_env_vars_cmd(connection, args):
    env_vars_info = connection.get_env_vars()
    print_item(env_vars_info, args)


def put_env_vars_cmd(connection, args):
    env_vars_path = args.env_vars_file

    if not os.path.exists(env_vars_path):
        exit_with_error("The specified path '%s' doesn't exist!" % (env_vars_path,))
    with open(env_vars_path, "r", encoding="utf-8") as env_vars_content:
        env_vars_info = connection.put_env_vars(json.loads(env_vars_content.read()))
        print_item(env_vars_info, args)


def post_env_vars_cmd(connection, args):
    env_vars_path = args.env_vars_file

    if not os.path.exists(env_vars_path):
        exit_with_error("The specified path '%s' doesn't exist!" % (env_vars_path,))
    with open(env_vars_path, "r", encoding="utf-8") as env_vars_content:
        env_vars_info = connection.post_env_vars(json.loads(env_vars_content.read()))
        print_item(env_vars_info, args)


def delete_env_var_cmd(connection, args):
    env_vars_info = connection.delete_env_var(args.env_var)
    print_item(env_vars_info, args)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import configparser
import os.path

import appdirs


def get_configfilename():
    data_dir = appdirs.user_data_dir(appname="sesamclient", appauthor="Sesam")
    configfilename = os.path.join(data_dir, "settings.ini")
    return configfilename


def get_configparser():
    myconfigparser = configparser.ConfigParser()
    myconfigparser.read(get_configfilename(), encoding="utf-8")
    return myconfigparser


def save_configparser_to_file(myconfigparser):
    configfilename = get_configfilename()
    data_dir = os.path.dirname(configfilename)
    os.makedirs(data_dir, exist_ok=True)
    with open(configfilename, "w", encoding="utf-8") as configfile:
        myconfigparser.write(configfile)


def config_cmd(args):
    """This function handles the 'config' command.

    Note that we don't have to check if the args.optionname is valid here, since that check has already been
    performed by the argparser.

    :param args: The arguments to the config-command.
    """
    if "." not in args.optionname:
        raise AssertionError("config key does not contain a section: %s" % (args.optionname,))

    section, optionname = args.optionname.split(".", maxsplit=1)

    myconfigparser = get_configparser()
    if args.optionvalue:
        # set the config value
        optionvalue = " ".join(args.optionvalue)
        if not myconfigparser.has_section(section):
            myconfigparser.add_section(section)
        myconfigparser.set(section, optionname, optionvalue)

        save_configparser_to_file(myconfigparser)
    else:
        # get the configvalue
        value = myconfigparser.get(section, optionname, fallback="")
        if value:
            print(value)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import pprint


class PumpIsAlreadyRunning(Exception):
    def __init__(self, running_task):
        self.running_task = running_task


class PipeAlreadyExists(Exception):
    pass


class SystemAlreadyExists(Exception):
    pass


class TimeoutWhileWaitingForRunningPumpToFinishException(Exception):
    pass


class TimeoutWhileWaitingForRunningIndexToFinishUpdatingException(Exception):
    pass


class TimeoutWhileWaitingForBackupToFinishException(Exception):
    pass

class TimeoutWhileWaitingForRestoreToFinishException(Exception):
    pass

class ExceptionWithResponse(Exception):
    def __init__(self, msg=None, response=None):
        super_args = []
        if msg:
            super_args.append(msg)
        super().__init__(*super_args)
        self.response = response


class InternalServerError(ExceptionWithResponse):
    """This is raised when the server has responded with a 5xx statuscode"""
    pass


class ServiceUnavailable(InternalServerError):
    """This is raised when the server has responded with a 503 statuscode"""
    pass


class BadRequestException(ExceptionWithResponse):
    """This is raised when the server has responded with an 4xx statuscode"""
    pass


class ForbiddenException(BadRequestException):
    """This is raised when the server has responded with an 403 statuscode"""
    pass


class PumpDoesNotSupportTheOperation(BadRequestException):
    def __init__(self, pump, operation, response):
        super().__init__(response=response)
        self.pump = pump
        self.operation = operation
        assert response is not None


class ConfigUploadFailed(BadRequestException):
    """This is raised when the server has responded with a 400 statuscode on a configuration upload request. The
    response object is stored in the 'response' attribute. The json payload of the response is stored in the
    'parsed_response' attribute."""
    def __init__(self, msg=None, response=None):
        if response is not None:
            from .utils import parse_json_response
            self.parsed_response = parse_json_response(response)
        else:
            self.parsed_response = None

        if msg is None:
            msg = "Config upload failed! response:\n%s" % (pprint.pformat(self.parsed_response),)

        super().__init__(msg=msg, response=response)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import json
import logging
import copy

from .exceptions import ConfigUploadFailed
from .entitybase import EntityBase
from . import utils


logger = logging.getLogger(__name__)


class System(EntityBase):
    """
    This class represents a system.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def delete(self, **kwargs):
        """Deletes this system from the sesam-node.
        """
        url = self._connection.get_system_url(self.id)
        response = self._connection.do_delete_request(url, allowable_response_status_codes=[200], **kwargs)

        self.update_raw_jsondata(response.content.decode("utf-8"))

    def modify(self, system_config, force=False, **kwargs):
        """Modifies the system with the specified configuration."""
        request_params = {"force": "true" if force else "false"}

        system_config = copy.deepcopy(system_config)
        system_config["_id"] = self.id
        response = self._connection.do_put_request(self._connection.get_system_config_url(self.id),
                                                   allowable_response_status_codes=[200, 400],
                                                   json=system_config,
                                                   params=request_params, **kwargs)

        if response.status_code == 400:
            raise ConfigUploadFailed(response=response)

        self.update_raw_jsondata(response.content.decode("utf-8"))

    @property
    def runtime(self):
        return copy.deepcopy(self._raw_jsondata["runtime"])

    def get_source_prototypes(self, **kwargs):
        url = self._connection.get_system_source_prototypes_url(self.id)
        response = self._connection.do_get_request(url, allowable_response_status_codes=[200], **kwargs)
        return response.json()

    def get_sink_prototypes(self, **kwargs):
        url = self._connection.get_system_sink_prototypes_url(self.id)
        response = self._connection.do_get_request(url, allowable_response_status_codes=[200], **kwargs)
        return response.json()

    def get_secrets(self, include_internal_secrets=False, **kwargs):
        """This returns a list with the names of the secrets."""
        request_params = {}
        if include_internal_secrets:
            request_params["include-internal-secrets"] = "true"

        response = self._connection.do_get_request(self._connection.get_secrets_url(system_id=self.id),
                                                   allowable_response_status_codes=[200],
                                                   params=request_params, **kwargs)
        secrets_info = utils.parse_json_response(response)
        return secrets_info

    def get_secrets_values(self, **kwargs):
        """This returns a dict with secretname=>secretvalue mappings. This is only available for nodes that
        are running in ci-test mode. In addition, the user has to have the "SystemAdmin" role."""
        request_params = {}
        response = self._connection.do_get_request(self._connection.get_secrets_values_url(system_id=self.id),
                                                   allowable_response_status_codes=[200],
                                                   params=request_params, **kwargs)
        secrets_info = utils.parse_json_response(response)
        return secrets_info

    def put_secrets(self, secrets_content, dont_encrypt=False, **kwargs):
        encrypt = dont_encrypt and "false" or "true"
        response = self._connection.do_put_request(self._connection.get_secrets_url(system_id=self.id),
                                                   allowable_response_status_codes=[200],
                                                   json=secrets_content, params=dict(encrypt=encrypt), **kwargs)

        secrets_info = utils.parse_json_response(response)
        return secrets_info

    def post_secrets(self, secrets_content, dont_encrypt=False, **kwargs):
        encrypt = dont_encrypt and "false" or "true"
        response = self._connection.do_post_request(self._connection.get_secrets_url(system_id=self.id),
                                                    allowable_response_status_codes=[200],
                                                    json=secrets_content, params=dict(encrypt=encrypt), **kwargs)

        secrets_info = utils.parse_json_response(response)
        return secrets_info

    def delete_secret(self, key, **kwargs):
        response = self._connection.do_delete_request(self._connection.get_secret_url(key, system_id=self.id),
                                                      allowable_response_status_codes=[200], **kwargs)
        secrets_info = utils.parse_json_response(response)

        return secrets_info

    def get_status(self, **kwargs):
        """This returns a dict with the system's status information."""
        request_params = {}
        response = self._connection.do_get_request(self._connection.get_system_status_url(self.id),
                                                   allowable_response_status_codes=[200],
                                                   params=request_params, **kwargs)
        status_info = utils.parse_json_response(response)

        return status_info

    @property
    def is_valid_config(self):
        return self._raw_jsondata["runtime"]["is-valid-config"]

    @property
    def config_errors(self):
        return self._raw_jsondata["runtime"]["config-errors"]



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import json
import sys
import textwrap

import yaml
import os.path

from .. import entity_json


def print_message(message, file=None, indent=0):
    if message:
        if not message.endswith("\n"):
            message += "\n"
        if file is None:
            file = sys.stdout
        if indent:
            index_prefix = " " * indent
            message = textwrap.indent(message, index_prefix)
        file.write(message)
        file.flush()


def _exit(status=0, message=None):
    if message:
        print_message(message, sys.stderr)

    sys.exit(status)


def exit_with_error(message):
    if message:
        message = "error: " + message
    _exit(status=1, message=message)


def item_as_json_object(item):
    if isinstance(item, (str, int, float)):
        return item
    return item.as_json_object()


def print_item_list(itemlist, args):

    ids_only = getattr(args, "ids_only", False)
    if args.out == "json":
        if ids_only:
            json_object_list = [item.id for item in itemlist]
        else:
            json_object_list = [item_as_json_object(item) for item in itemlist]
        print_message(entity_json.entities_to_json(json_object_list, indent=4))

    elif args.out == "yaml":
        if ids_only:
            json_object_list = [item.id for item in itemlist]
        else:
            json_object_list = [item_as_json_object(item) for item in itemlist]

        print_message(yaml.dump(json_object_list))

    elif args.out == "plain":
        for item in itemlist:
            if ids_only:
                print_message(str(item.id))
            else:
                print_message(str(item))

    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))


def print_item(item, args):
    if isinstance(item, dict) or item is None:
        json_object = item
    else:
        json_object = item.as_json_object()

    if args.out == "json":
        print_message(entity_json.entities_to_json(json_object, indent=4))

    elif args.out == "yaml":
        print_message(yaml.dump(json_object))

    elif args.out == "plain":
        print_message(str(item))

    else:
        raise NotImplementedError("Output format '%s' has not been implemented!" % (args.out,))


def is_valid_configfile_name(filename):
    filename = os.path.basename(filename)
    return filename.endswith(".conf.json") or is_valid_but_deprecated_configfile_name(filename)


def is_valid_but_deprecated_configfile_name(filename):
    filename = os.path.basename(filename)
    return filename == "nodeconfig.json"
